import argparse
import configparser
import ast
import os
import platform
import sys
import importlib
import traceback
import logging
from . import mytypes
from . import libfuncs
from .symboltable import symtab
from . import astutils
from . import utils


class UnsupportedException(Exception):
    def __init__(self,*args,**kwargs):
        Exception.__init__(self,*args,**kwargs)

class UnhandledNodeException(Exception):
    def __init__(self, N):
        Exception.__init__(self, ast.dump(N))

class UndefinedSymbolException(Exception):
    def __init__(self, module, funcname):
        self.module = module
        self.funcname = funcname
        Exception.__init__(self)

    def __str__(self):
        if not self.module:
            self.module = 'built-in'
        s = 'undefined function `%s` in module `%s`' % (self.funcname, self.module)
        return s    
        
# class UnhandledNodeException(Exception):
#     def __init__(self, N):
#         Exception.__init__(self, ast.dump(N))

# def dump(N):
#     print(ast.dump(N))
                

def dump(N, msg=''):
    print(msg, ast.dump(N))
                

args = None
options = {}
func_versions = {}

def init():
    '''
    Set up global stuff
    '''
    global args, options, func_versions
    options = {}
    func_versions = {}
    parse_args()
    setup_loggers()


def setup_loggers():
    # l = utils.setup_logger('ea', 'ea.log')
    # if args.verbose:
    #     l.setLevel(logging.DEBUG)

    l = utils.setup_logger('type', 'type.log')
    if args.dump_type:
        l.setLevel(logging.DEBUG)

    
        
def parse_args():
    global args, options, func_versions
    desc ='''Compiles an Intrepydd program into native code that can be imported as a Python module.
    The command-line options below are for experimental features under development.
    '''
    parser = argparse.ArgumentParser(description=desc)        
    parser.add_argument('file', help='input file')
    # parser.add_argument('--pymodule', default=True,
    #                     action='store_true', help='produce a python module')
    parser.add_argument('-host', default='py', help='[py|cpp]')
    parser.add_argument('-I', default='', help='Header file search location, separated by `:`')
    parser.add_argument('-compiler-flags', default='', help='Compiler flags, separated by `:`')
    parser.add_argument('-linker-flags', default='', help='Linker flags, separated by `:`')
    parser.add_argument('-func-version', default='', help='Specify the version number of a function')
    parser.add_argument('-opt-level', default=2, help='Optimization level, defaults to 2')
    parser.add_argument('-O0', default=False, action='store_true', help='Generate Python code (in progress)')
    parser.add_argument('-O1', default=False, action='store_true', help='Generate Numba code (in progress)')
    parser.add_argument('-O2', default=False, action='store_true', help='Generate native code (default)')
    parser.add_argument('-g', default=False, action='store_true', help='Generate code with debug info')
    parser.add_argument('-dumppy', default=False, action='store_true', help='Dump Python code from optimized AST via O2')
    parser.add_argument('-blas', default=True, action='store_true', help='Use BLAS calls')
    parser.add_argument('-verbose', default=False, action='store_true', help='General verbose flag')
    parser.add_argument('-cpp', default=False, action='store_true', help='Set host to cpp')
    parser.add_argument('-no-compile', default=False, action='store_true', help='Just generate cpp file but not compile it')
#    parser.add_argument('-g', default=False, action='store_true', help='Enable debug mode')
    parser.add_argument('-unchecked-access', default=True, action='store_true', help='Fast but unchecked array access')
    parser.add_argument('-licm', default=False, action='store_true', help='Loop Invariant Code Motion')
    parser.add_argument('-licm-array', default=False, action='store_true', help='Loop Invariant Code Motion for array references')
    parser.add_argument('-licm-ast', default=False, action='store_true', help='Loop Invariant Code Motion for AST-level expressions (general)')
    parser.add_argument('-array-opt', default=False, action='store_true', help='Sparse/dense array optimizations')
    parser.add_argument('-sparse-opt', default=False, action='store_true', help='Sparse array optimizations')
    parser.add_argument('-dense-opt', default=False, action='store_true', help='Dense array optimizations')
    parser.add_argument('-slice-opt', default=False, action='store_true', help='Use raw pointers for slices')
    # Tong: This option is potentially buggy, disable for now
    # parser.add_argument('-alloc-opt', default=False, action='store_true', help='Array allocation optimizations via escape analsis')

    parser.add_argument('-dump-type', default='', help='Dump type info for locals (defaults to all locals). Possible values: "*" (all), "!__" (no __xx vars), "[var_name]"')
    
    args = parser.parse_args()

    if args.O2:
        args.opt_level = 2
    elif args.O1:
        args.opt_level = 1
    elif args.O0:
        args.opt_level = 0
    else:
        args.opt_level = 2

    if args.cpp:
        args.host = 'cpp'
    if args.func_version:    
        funcs = args.func_version.split(',')
        for f in funcs:
            items = f.split(':')
            funcname = items[0]
            version = items[1]
            func_versions[funcname] = version

    if args.dumppy:
        options['dumppy'] = True

    if args.licm:
        options['licm_array'] = True
        options['licm_ast'] = True

    if args.licm_array:
        options['licm_array'] = True

    if args.licm_ast:
        options['licm_ast'] = True

    if args.array_opt:
        options['sparse_opt'] = True
        options['dense_opt'] = True
    if args.sparse_opt:
        options['sparse_opt'] = True
    if args.dense_opt:
        options['dense_opt'] = True

    #print(args)

def get_func_version(name):
    return func_versions.get(name, 0)

def parse_config_file():
    global options
    C = configparser.ConfigParser()
    C.read("config.ini")
    for section in C.sections():
        opts = C.options(section)
        print(opts)
        for opt in opts:
            options[opt] = C.get(section, opt)
    
    print(options)
            
def get_option(o):
    return options[o]

cpp_ofs = None

def open_cpp_ofs(mod_name):
    global cpp_ofs
    cpp_ofs = open(mod_name+'.cpp', 'w')

def close_cpp_ofs(mod_name):
    global cpp_ofs
    cpp_ofs.close()

class CppModule(object):
    def __init__(self, name):
        self.ofs = open(name+'.cpp', 'w')
        self.pymodule = False
        self.funcs = []
        self.imports = {}
        if args.opt_level == 2:
            self.pymodule = True

    def close(self):
        self.ofs.close()      

    def writeln(self, s=''):
        print(s, file=self.ofs)

    def write(self, s):
        print(s, file=self.ofs, end='')

    def add_function(self, N):
        self.funcs.append(N)

    def add_imported_module(self, alias):
        realname = alias.name
        asname = alias.asname
        cppmodname = realname
        if is_imported_module_standard(cppmodname):
            cppmodname = 'std'            
        
        if asname:
            self.imports[asname] = cppmodname
        else:
            self.imports[realname] = cppmodname

    def get_module_namespace(self, name):
        return self.imports[name]

    def is_module(self, name):
        return name in self.imports

    def do_pymodule_header(self):
        if not self.pymodule:
            return


        cxxflags = ['-std=c++11']
        if platform.system() == 'Darwin':
            cxxflags += ['-stdlib=libc++']
        for p in get_include_paths():
            cxxflags.append('-I' + p)
        if args.I:
            for p in args.I.split(':'):
                cxxflags.append('-I' + p)

        cxxflags += libfuncs.get_compiler_flags()
        if args.compiler_flags:    
            cxxflags += args.compiler_flags.split(':')
        cfg = '<%\n'
        cfg += "cfg['compiler_args'] = %s\n" % str(cxxflags)

        linker_flags = libfuncs.get_linker_flags()

        if args.linker_flags:    
            linker_flags += args.linker_flags.split(':')
        cfg += "cfg['linker_args'] = %s\n" % str(linker_flags)
        cfg += "setup_pybind11(cfg)\n"
        cfg += '%>'
        self.writeln(cfg)
        self.writeln('')

    def do_default_include(self):
        self.writeln('#include "shared/rt.hpp"')
        if args.host == 'cpp':
            self.writeln('#include "cpp/NdArray.hpp"')
        elif args.host == 'py':
            self.writeln('#include <pybind11/pybind11.h>')
            self.writeln('#include <pybind11/stl.h>')
            self.writeln('#include <pybind11/numpy.h>')
        else:
            assert False
        
        for h in libfuncs.get_header_files():
            self.writeln('#include "%s"' % h)
            
        self.writeln('')

        if args.host == 'py':    
            self.writeln('namespace py = pybind11;\n')


    def do_py_footer(self):
        if not self.pymodule:
            return
        m = get_module_name()
        self.writeln('PYBIND11_PLUGIN(%s) {' % m)
        self.writeln('  pybind11::module m("%s", "auto-compiled c++ extension");' % m)
        for F in self.funcs:
            fn = F.name
            s = '  m.def("%s", &%s::%s,' % (fn, m, fn)
            for arg in F.args.args:
                s += 'py::arg().noconvert(),'
            self.writeln(s[:-1] + ');')
        self.writeln('  return m.ptr();')
        self.writeln('}')  

def is_imported_module_standard(name):
    if name == 'cmath':
        return True
    return False

    
cpp_module = None
        
def open_cpp_module(name):
    global cpp_module
    cpp_module = CppModule(name)

def close_cpp_module():
    cpp_module.close()
        
def get_basefile():
    return os.path.basename(args.file)

def get_filepath():
    d = os.path.dirname(args.file)
    if d:
        return d + '/'
    else:
        return './'

def get_module_name():
    return os.path.basename(get_module_fullname())

def get_module_fullname():
    return args.file.replace('.pydd', '').replace('.py', '').replace('.cpp', '')

def get_module_path(s):
    return os.path.dirname(get_module_fullname())
    
def get_include_paths():
    paths = []
    paths.append(get_compiler_path() + '/include/')
    paths.append(get_compiler_path() + '/library/')
    return paths

def get_compiler_path():
    return os.path.dirname(os.path.realpath(__file__))

    
def is_user_defined_func(name):
    # Assume all non-user-defined functions are built-in
    return symtab.is_user_defined_symbol(name)

def get_module_func_ret_type(module, func, arg_types):
    path = "%s/library/%s/metadata.py" % (get_compiler_path(), module)
    spec = importlib.util.spec_from_file_location('metadata', path)
    M = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(M)
    funcinfo = M.funcinfo
    
def get_imported_module_info(module):
    path = "%s/library/%s/metadata.py" % (get_compiler_path(), module)
    spec = importlib.util.spec_from_file_location('metadata', path)
    M = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(M)
    return M
    

def import_file(full_name, path):
    """Import a python module from a path. 3.4+ only.

    Does not call sys.modules[full_name] = path
    """
    spec = importlib.util.spec_from_file_location(full_name, path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod


unrecognized_type_annotation = 'unrecognized type annotation'
unsupported_node = 'unsupported node'

def exit_on_unsupported_node(N):
    exit_on_node(N, 'unsupported node')

def exit_on_unsupported_type_ann(N):
    exit_on_node(N, 'unrecognized type annotation')

def exit_on_node(N, msg=""):
#    print('node', vars(N))
    if args.verbose:
        traceback.print_stack()
        print('=================== Stack Trace Ends =================')
    print('error:', msg)
    print('problematic node:', ast.dump(N))
    if astutils.has_location(N):
        print('line number:', astutils.get_lineno(N))
        print('column offset:', N.col_offset)
        print('problematic line:\n')
        lineno = 0
        for line in open(args.file):
            lineno += 1
            if lineno == N.lineno:
                print(line)
        for i in range(N.col_offset):
            print(' ', end='')
        print('^')    
    sys.exit(1)



from enum import Enum
import ast
from . import libfuncs
from . import defuse

'''
Support scalar privatization for loop parallelization.

Todo:
 - Standardize options, print (verbose and warning), and assert
'''

### Options ###
# 0: Nothing, 1: Warning, 2: Debug (light), 3: Debug (moderate), 4: Debug (heavy)
verbose_level = 1
# To enable nested parallel loops
enable_nested_pfor = False

### Functions to access/modify ast.For nodes ###
def is_pfor(N: ast.For):
    return N.is_pfor

def serialize_pfor(N: ast.For, reason):
    N.is_pfor = False

    global verbose_level
    if verbose_level >= 1:
        print('[Warning] Serialize pfor at line', N.lineno, 'due to:', reason)

def add_private_sets(N: ast.For, vars_private, vars_lastprivate):
    assert N.is_pfor
    N.vars_private = vars_private
    N.vars_lastprivate = vars_lastprivate

    global verbose_level
    if verbose_level >= 2:
        ln = N.lineno if hasattr(N, 'lineno') else '???'
        print('[Verbose] Safe pfor at line', ln,
              'with private:', vars_private, '/ lastprivate:', vars_lastprivate)

is_first_to_enable_openmp = False
def enable_openmp():
    global is_first_to_enable_openmp
    if not is_first_to_enable_openmp:
        libfuncs.add_function_module('omp')
        is_first_to_enable_openmp = True

### Utilities ###
def to_csv(vars):
    s = ''
    if vars:
        s += vars.pop()
        for var in vars:
            s += ', %s' % var
    return s


############################################################################################

class ScalarPrivatization(defuse.DefUseAnalysis):

    def __init__(self):
        '''
        Information per Function Definition
        '''
        defuse.DefUseAnalysis.__init__(self)
        self.nodes_pfor = []  # LevelTreeNode list of pfor loops

    def verify_fields(self):
        defuse.DefUseAnalysis.verify_fields(self)
        assert not self.nodes_pfor

    def clear_level_tree_information(self):
        defuse.DefUseAnalysis.clear_level_tree_information(self)
        self.nodes_pfor.clear()

    def add_nodes_pfor(self, level_node, N: ast.For):
        if is_pfor(N):
            self.nodes_pfor.append(level_node)

    def serialize_enclosing_pfors(self, reason):
        curr = level_tree_current
        while curr:
            if curr.prev:
                curr = curr.prev
            else:
                curr = curr.parent
                if curr and (curr.kind is self.LevelTreeNode.Kind.FOR):
                    N = curr.ast_node
                    assert isinstance(N, ast.For)
                    if is_pfor(N):
                        serialize_pfor(N, reason)
                        self.nodes_pfor.remove(curr)
    '''
    Note:
    This algorithm to compute live-out set is not exact, i.e., the return value
    is set of variables that 'may' live out from the pfor specified by level_pfor.
    However, this decision is always safe in terms of privatization.
    '''
    def compute_vars_live_out(self, vars_target: set, level_pfor):
        vars_live_out = set()
        for var in vars_target:
            if var not in self.var_to_nodes:
                continue
            is_live_out = False
            for node_ref in self.var_to_nodes[var]:
                if defuse.vector_include(level_pfor, node_ref.level):
                    continue
                node_s = None
                node_t = node_ref
                curr = node_ref
                while curr:
                    if curr.prev:
                        if var in curr.prev.vars_def:
                            node_s = curr.prev
                            break
                        curr = curr.prev
                    else:
                        curr = curr.parent
                        if curr and (curr.kind is self.LevelTreeNode.Kind.FOR):
                            assert isinstance(curr.ast_node, ast.For)
                            node_t = curr
                if not node_s:
                    node_s = self.level_tree_head

                if (defuse.vector_compare(level_pfor, node_s.level) >= 0
                    and defuse.vector_compare(node_t.level, level_pfor) >= 0):
                    is_live_out = True
                    break
            if is_live_out:
                vars_live_out.add(var)
        return vars_live_out

    def get_outer_pfor(self, level_node):
        curr = level_node
        while curr:
            if curr.prev:
                curr = curr.prev
            else:
                curr = curr.parent
                if curr and (curr.kind is self.LevelTreeNode.Kind.FOR):
                    N = curr.ast_node
                    assert isinstance(N, ast.For)
                    if is_pfor(N):
                        return curr  # Return outer pfor
        return None

    def eval_pfor(self, level_node):
        N = level_node.ast_node
        assert is_pfor(N)

        is_parallel = False
        level = level_node.level
        if (not enable_nested_pfor) and self.get_outer_pfor(level_node):
            serialize_pfor(N, 'outer pfor (no nested pfor supported).')
        elif level_node.vars_use & (level_node.vars_may_def | level_node.vars_cond_def):
            serialize_pfor(N, 'exposed use.')
        elif self.compute_vars_live_out(level_node.vars_may_def, level):
            serialize_pfor(N, 'live-ouf of may-def variables.')
        else:
            is_parallel = True
            vars_private = level_node.vars_may_def
            vars_lastprivate = self.compute_vars_live_out(level_node.vars_cond_def, level)
            vars_private |= (level_node.vars_cond_def - vars_lastprivate)
            add_private_sets(N, to_csv(vars_private), to_csv(vars_lastprivate))
            enable_openmp()
        return is_parallel

    def eval_all_pfors(self):
        global verbose_level
        if verbose_level >= 3:
            self.show_fields(verbose_level >= 4)

        updated = []
        for node in self.nodes_pfor:
            if self.eval_pfor(node):
                updated.append(node)
        self.nodes_pfor = updated

        if verbose_level >= 3:
            print('')


############################################################################################

    def visit_FunctionDef(self, N: ast.FunctionDef):
        if N.has_pfor:
            self.verify_fields()

            self.visit_FunctionDef_main(N)

            self.eval_all_pfors()

            self.clear_level_tree_information()

    def visit_For_head(self, N: ast.For):
        for_node = self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.FOR)
        self.add_nodes_pfor(for_node, N)
        self.visit(N.iter)

        self.update_level_tree_with_child(N.target, self.LevelTreeNode.Kind.FOR_IN)
        if isinstance(N.target, ast.Name):
            self.define_var(N.target.id)
        else:
            # Loop iterator is not scalar
            self.visit(N.target)

        return for_node

    def visit_Continue(self, node):
        self.serialize_enclosing_pfors('continue')

    def visit_Break(self, node):
        self.serialize_enclosing_pfors('break')

    def visit_Assert(self, node):
        self.serialize_enclosing_pfors('assert')



from enum import Enum
import ast
from . import libfuncs
from . import mytypes

'''
Support def use analysis.

Todo:
 - Element-wise analysis
 - Standardize options, print (verbose and warning), and assert
'''

# Fuctions whose arguments are read-only
a1l = 'arg1_len'
adm = 'arg_max_dim'
rsm = 'red_sum'
mml = 'matmult'
pure_funcs = {
    'len': 0, 'boolean': 0, 'int32': 0, 'int64': 0, 'float32': 0, 'float64': 0, 'range': 1,
    'zeros': a1l, 'zeros_2d': 2, 'empty': a1l, 'empty_2d': 2,
    'empty_like': adm, 'get_row': 1, 'get_col': 1, 'Array': 0,
    'sum_rows': 1, 'shape': 0, 'stride': 0, 'dsyrk': 2,
    # Reduction
    'sum': rsm, 'prod': 0, 'min': 0, 'max': 0, 'argmin': 0, 'argmax': 0,
    'any': 0, 'all': 0, 'allclose': 0, 'where': 1,
    # Element-wise unary
    'abs': adm, 'minus': adm, 'isnan': adm, 'isinf': adm, 'elemwise_not': adm, 'sqrt': adm,
    'exp': adm, 'cos': adm, 'sin': adm, 'tan': adm, 'acos': adm, 'asin': adm, 'atan': adm,
    # Element-wise binary
    'add': adm, 'sub': adm, 'mul': adm, 'maximum': adm, 'pow': adm, 'div': adm, 'log': adm,
    'eq': adm, 'neq': adm, 'lt': adm, 'gt': adm, 'le': adm, 'ge': adm,
    'logical_and': adm, 'logical_or': adm, 'logical_xor': adm, 'logical_not': adm,
    # Matrix operations
    'transpose': 2, 'innerprod': 0, 'matmult': mml,
    'lu': 2, 'qr': 2, 'eig': 2, 'svd': 2, 'tril': 2, 'triu': 2, 'diag': 2, 'kron': 2, 'convolve': 2,
    # Sparse matrix operations
    'empty_spm': 2, 'csr_to_spm': 2, 'arr_to_spm': 2,
    'getval': 0, 'getnnz': 0, 'getcol': 0, 'nnz': 0,
    'spm_add': 2, 'spm_mul': 2, 'spmm': 2, 'spmv': 1,
    'spmm_dense': 2, 'sparse_diags': 2, 'spm_diags': 2,
    # Heap
    'heapinit': 1, 'heappeek': 0, 'heapsize': 0,
}

upd_funcs = {
    'append': 0, 'set_row': 0, 'set_col': 0, 'plus_eq': 0, 'minus_eq': 0,
    # Sparse matrix operations
    'spm_to_csr': [1,2,3], 'spm_set_item': 0, 'spm_set_item_unsafe': 0,
    # Heap
    'heappush': 0, 'heappop': 0,
}

### Utilities ###
def vector_include(vect1, vect2):
    n = len(vect1)
    if len(vect2) <= n:
        return False
    for i in range(n):
        if vect1[i] != vect2[i]:
            return False
    return True

def vector_compare(vect1, vect2):
    n = len(vect1) if len(vect1) < len(vect2) else len(vect2)
    for i in range(n):
        if (vect1[i] < vect2[i]):
            return -1
        elif (vect1[i] > vect2[i]):
            return 1
    return 0


############################################################################################

class DefUseAnalysis(ast.NodeVisitor):
    '''
    Information per AST node
    '''
    class LevelTreeNode:
        class Kind(Enum):
            UNDEF = -1
            FUNC = 0
            FOR = 1
            FOR_IN = 2
            WHILE = 3
            WHILE_IN = 4
            IF = 5
            IF_IN = 6
            ELSE = 7
            ELSE_IN = 8
            ASSIGN = 9
            AUG_ASSIGN = 10
            ANN_ASSIGN = 11
            CALL = 12
            RETURN = 13

        def __init__(self, level, ast_node, kind, prev, parent):
            if isinstance(level, list):
                self.level = tuple(level)
            else:
                self.level = level
            self.ast_node = ast_node
            self.kind = kind
            # Variable sets
            self.vars_use = set()       # Exposed use (not self-defined value)
            self.vars_def = set()       # Must def
            self.vars_cond_def = set()  # Must def within body (If/Else/Loop)
            self.vars_may_def = set()
            self.vars_em_def = set()
            # Tree link
            self.prev = prev
            self.next = None
            self.parent = parent
            self.child = None

        def show(self):
            print('[Verbose] dump LevelTreeNode of level:', self.level, '-', self.kind)
            if self.vars_use: print(' - vars_use', self.vars_use)
            if self.vars_def: print(' - vars_def', self.vars_def)
            if self.vars_may_def: print(' - vars_may_def', self.vars_may_def)
            if self.vars_em_def: print(' - vars_em_def', self.vars_em_def)
            if self.vars_cond_def: print(' - vars_cond_def', self.vars_cond_def)

        def copy(self):
            cp = DefUseAnalysis.LevelTreeNode(self.level, self.ast_node, self.kind,
                                              self.prev, self.parent)
            cp.vars_use = self.vars_use.copy()
            cp.vars_def = self.vars_def.copy()
            cp.vars_may_def = self.vars_may_def.copy()
            cp.vars_em_def = self.vars_em_def.copy()
            cp.vars_cond_def = self.vars_cond_def.copy()
            cp.next = self.next
            cp.child = self.child
            return cp

        def is_assign(self):
            return (self.kind is DefUseAnalysis.LevelTreeNode.Kind.ASSIGN
                    or self.kind is DefUseAnalysis.LevelTreeNode.Kind.AUG_ASSIGN
                    or self.kind is DefUseAnalysis.LevelTreeNode.Kind.ANN_ASSIGN)

        def update_child_level(self):
            ch = self.child
            nl = list(self.level)
            nl.append(0)
            while ch:
                nl[-1] = ch.level[-1]
                ch.level = tuple(nl)
                if ch.child:
                    ch.update_child_level()
                ch = ch.next

    def __init__(self):
        '''
        Information per Function Definition
        '''
        self.curr_func = None
        self.level_tree_head = None
        self.level_tree_current = None
        self.status = []        # Current level during traverse
        self.var_to_nodes = {}  # Key: variable, Value: LevelTreeNodes to ref variable

    def verify_fields(self):
        assert not self.level_tree_head
        assert not self.level_tree_current
        assert not self.status
        assert not self.var_to_nodes

    def show_fields(self, show_all, curr = None):
        is_top = False
        if curr is None:
            curr = self.level_tree_head
            is_top = True

        if show_all or (curr.kind is self.LevelTreeNode.Kind.FOR
                        or curr.kind is self.LevelTreeNode.Kind.WHILE):
            curr.show()
        if curr.child:
            self.show_fields(show_all, curr.child)
        if curr.next:
            self.show_fields(show_all, curr.next)

        if is_top:
            print("[Verbose] dump mapping: variable -> levels to use\n", end = ' - ')
            for var in self.var_to_nodes:
                print(var, ':', end = ' ')
                for node in self.var_to_nodes[var]:
                    print(node.level, end = ', ')
                print(end = ' ')
            print('')

    def delete_tree_branches(self, curr):
        next = curr.next
        if next:
            next.prev = None
            curr.next = None
            self.delete_tree_branches(next)
            del next
        child = curr.child
        if child:
            curr.child = None
            child.parent = None
            self.delete_tree_branches(child)
            del child
        curr = None

    def clear_level_tree_information(self):
        self.var_to_nodes.clear()
        self.status.clear()

        tmp = self.level_tree_head
        self.level_tree_head = None
        self.level_tree_current = None
        if tmp:
            self.delete_tree_branches(tmp)
            del tmp

    def define_var(self, var):
        self.level_tree_current.vars_def.add(var)

    def define_em_var(self, var):
        self.level_tree_current.vars_em_def.add(var)

    def define_em_var_recurse(self, N):
        if isinstance(N, ast.Name) and N.id in self.curr_func.typemap:
            ty = self.curr_func.typemap[N.id]
            assert ty
            if isinstance(ty, mytypes.IterableType):
                self.define_em_var(N.id)

        for node in ast.iter_child_nodes(N):
            self.define_em_var_recurse(node)

    def check_and_def_em_subscr(self, N):
        if isinstance(N, ast.Subscript) and isinstance(N.value, ast.Name):
            self.define_em_var(N.value.id)

    def check_and_def_em_func(self, N):
        if not isinstance(N, ast.Call):
            return

        args = N.args
        if isinstance(N.func, ast.Name):
            func = N.func.id
            if func in pure_funcs:
                return
            elif func in upd_funcs:
                upd = upd_funcs[func]
                if isinstance(upd, int):
                    self.define_em_var_recurse(args[upd])
                else:
                    assert isinstance(upd, list)
                    for u in upd:
                        if u < len(args):
                            self.define_em_var_recurse(args[u])
                return

        for arg in args:
            self.define_em_var_recurse(arg)

    def use_var(self, var):
        if var not in self.level_tree_current.vars_use:
            self.level_tree_current.vars_use.add(var)
            if var in self.var_to_nodes:
                self.var_to_nodes[var].add(self.level_tree_current)
            else:
                self.var_to_nodes[var] = {self.level_tree_current}

    def add_level_tree_head(self, N, kind):
        self.status = [0]
        self.level_tree_head = self.LevelTreeNode(self.status, N, kind, None, None)
        self.level_tree_current = self.level_tree_head

    def update_level_tree_with_sibling(self, N, kind):
        self.status[-1] += 1
        prev = self.level_tree_current
        next = self.LevelTreeNode(self.status, N, kind, prev, None)
        prev.next = next
        self.level_tree_current = next
        return next

    def update_level_tree_with_child(self, N, kind):
        self.status.append(0)
        parent = self.level_tree_current
        child = self.LevelTreeNode(self.status, N, kind, None, parent)
        parent.child = child
        self.level_tree_current = child
        return child

    def collect_child_def_use(self, level_node):
        vars_use = set()
        vars_def = set()
        vars_may_def = set()
        vars_em_def = set()
        node = level_node.child
        while node:
            vars_use |= (node.vars_use - vars_def)
            vars_def |= node.vars_def
            vars_may_def |= (node.vars_may_def | node.vars_cond_def)
            vars_em_def |= node.vars_em_def
            node = node.next
        vars_may_def -= vars_def
        vars_em_def -= vars_def
        level_node.vars_use |= vars_use
        level_node.vars_cond_def |= vars_def
        level_node.vars_may_def |= vars_may_def
        level_node.vars_em_def |= vars_em_def

        self.status.pop()
        assert tuple(self.status) == level_node.level
        self.level_tree_current = level_node


############################################################################################

    def visit_body_core(self, N: ast.stmt):
        if isinstance(N, ast.Expr) and isinstance(N.value, ast.Call):
            self.visit_Call(N, True)
        else:
            self.visit(N)

    def visit_body(self, stmts: list):
        for stmt in stmts:
            self.visit_body_core(stmt)

    def visit_FunctionDef_main(self, N: ast.FunctionDef):
        self.add_level_tree_head(N, self.LevelTreeNode.Kind.FUNC)
        for item in N.args.args:
            self.visit(item)
            self.define_var(item.arg)
        if N.args.vararg is not None:
            self.visit(N.args.vararg)
            self.define_var(N.args.vararg.arg)
        for item in N.args.kwonlyargs:
            self.visit(item)
            self.define_var(item.arg)
        for item in N.args.kw_defaults:
            self.visit(item)
        if N.args.kwarg is not None:
            self.visit(N.args.kwarg)
            self.define_var(N.args.kwarg.arg)
        for item in N.args.defaults:
            self.visit(item)
        for item in N.decorator_list:
            self.visit(item)

        self.visit_body(N.body)

    def visit_FunctionDef(self, N: ast.FunctionDef):
        self.curr_func = N
        self.verify_fields()
        self.visit_FunctionDef_main(N)

        # Note: perform analysis here

        self.clear_level_tree_information()

    def visit_For_head(self, N: ast.For):
        for_node = self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.FOR)
        self.visit(N.iter)

        self.update_level_tree_with_child(N.target, self.LevelTreeNode.Kind.FOR_IN)
        if isinstance(N.target, ast.Name):
            self.define_var(N.target.id)
        else:
            # Loop iterator is not scalar
            self.visit(N.target)

        return for_node

    def visit_For(self, N: ast.For):
        for_node = self.visit_For_head(N)

        self.visit_body(N.body)
        self.collect_child_def_use(for_node)

        assert not N.orelse

    def visit_While_head(self, N: ast.While):
        while_node = self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.WHILE)
        self.visit(N.test)

        self.update_level_tree_with_child(None, self.LevelTreeNode.Kind.WHILE_IN)

        return while_node

    def visit_While(self, N: ast.While):
        while_node = self.visit_While_head(N)

        self.visit_body(N.body)
        self.collect_child_def_use(while_node)

        assert not N.orelse

    def visit_If(self, N: ast.If):
        if_node = self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.IF)
        self.visit(N.test)

        self.update_level_tree_with_child(None, self.LevelTreeNode.Kind.IF_IN)
        self.visit_body(N.body)
        self.collect_child_def_use(if_node)

        if N.orelse:
            else_node = self.update_level_tree_with_sibling(None, self.LevelTreeNode.Kind.ELSE)
            self.update_level_tree_with_child(None, self.LevelTreeNode.Kind.ELSE_IN)
            self.visit_body(N.orelse)
            self.collect_child_def_use(else_node)
            vars_def = if_node.vars_cond_def & else_node.vars_cond_def
            if vars_def:
                else_node.vars_def |= vars_def
                else_node.vars_cond_def -= vars_def

    def visit_Assign(self, N: ast.Assign):
        self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.ASSIGN)
        self.visit(N.value)
        for item in N.targets:
            if isinstance(item, ast.Name):
                self.define_var(item.id)
            else:
                # Assignment to non-scalar variable
                self.visit(item)
                self.check_and_def_em_subscr(item)

    def visit_AugAssign(self, N: ast.AugAssign):
        self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.AUG_ASSIGN)
        self.visit(N.value)
        if isinstance(N.target, ast.Name):
            self.use_var(N.target.id)
            self.define_var(N.target.id)
        else:
            # AugAssignment to non-scalar variable
            self.visit(N.target)
            self.check_and_def_em_subscr(N.target)

    def visit_AnnAssign(self, N: ast.AnnAssign):
        self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.ANN_ASSIGN)
        self.visit(N.annotation)
        self.visit(N.value)
        if isinstance(N.target, ast.Name):
            self.define_var(N.target.id)
        else:
            # AnnAssignment to non-scalar variable
            self.visit(N.target)
            self.check_and_def_em_subscr(N.target)

    def visit_Call(self, N, is_stmt = False):
        if is_stmt:
            assert isinstance(N, ast.Expr)
            self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.CALL)
            Nv = N.value
        else:
            Nv = N

        self.visit(Nv.func)
        for item in Nv.args:
            self.visit(item)
        for item in Nv.keywords:
            self.visit(item)

        if self.curr_func:
            self.check_and_def_em_func(Nv)

    def visit_Return(self, N: ast.Return):
        if N.value:
            self.update_level_tree_with_sibling(N, self.LevelTreeNode.Kind.RETURN)
            self.visit(N.value)

    def visit_Name(self, N: ast.Name):
        # modify to not use function names
        self.use_var(N.id)



import sys
import ast
from pprint import pprint
import traceback

from . import mytypes
from . import glb
from . import utils
from . import libfuncs
from .symboltable import symtab
from .glb import UnhandledNodeException
from .glb import dump
from .astutils import *


class AllocHoistTransformer(ast.NodeTransformer):
    def __init__(self):
        self.logger = utils.get_logger('ea')


    def visit_FunctionDef(self, F: ast.FunctionDef):
        body = F.body
        visited = set()
        for stmt in body:
            if is_loop(stmt):
                self.do_loop(stmt, body, visited)
        
        return F


    def do_loop(self, N, outerbody, visited):
        '''
        Each loop is associated with a set of allocation sites
        that are candidate for hoisting. This function removes
        these sites from the loop body and insert them immediately
        before the loop.
        '''
        if N in visited:
            return
        
        loopbody = N.body
        sites = N.alloc_sites
        
        for site in sites:    
            siteindex = loopbody.index(site)
            self.logger.debug("hoist site: %s" % site)
            arrayname = get_assign_target(site)
            fillcall = ast.Expr(create_call('fill', [arrayname, '0']))
            ast.copy_location(fillcall, site)
            # This will cause dumped python file line issue
            loopbody[siteindex] = fillcall

        #N.alloc_sites.clear()
        index = outerbody.index(N)

        sites.reverse()
        for site in sites:
            outerbody.insert(index, site)

        for stmt in loopbody:
            if is_loop(stmt):                
                self.do_loop(stmt, loopbody, visited)

        visited.add(N)





import intrepydd
from intrepydd.lang import *


def foo(max_iters: int32, a: Array(int32, 2), b: Array(int32, 2)):
    i = 0
    while i < max_iters:
        a += b + 1
        i += 1
    return a 

foo = intrepydd.compile(foo, licm=True)


# Built-in types and functions exposed to users

int64 = None
int32 = None
float64 = None
float32 = None

def Array(ty, ndim):
    return

def List(ty):
    return

def get_row(a, i):
    return a[i]




class MyType(object):
    def get_declare_str(self):
        if pass_by_ref(self):
            return self.get_cpp_name() + '*'
        else:
            return self.get_cpp_name()
        
    def equals(self, ty):
        if hasattr(self, 'etype') and hasattr(ty, 'etype'):
            return self.etype == ty.etype
        return self == ty

    # return true if types can be broadcasted
    # ex: (array<int>, int) -> true
    # ex: (array<double>, int) -> false
    # ex: (tuple<int>, int) -> false
    def can_broadcast_with(self, ty):
        if not (is_basic_type(self) or is_array(self)):
            return False
        if not (is_basic_type(ty) or is_array(ty)):
            return False
        e_type_1 = self if is_basic_type(self) else self.etype
        e_type_2 = ty if is_basic_type(ty) else ty.etype
        return e_type_1 == e_type_2

    
class IterableType(MyType):
    def get_elt_type(self):
        '''
        Returns the element type in a `for...in` clause
        '''
        return self.etype
        #raise NotImplementedError()

        
class ListType(IterableType):
    def __init__(self, etype):
        self.etype = etype
        self.initsize = 0
        
    def set_init_size(self, n):
        self.initsize = n

    def get_init_size(self):
        return self.initsize

    def get_cpp_name(self):
        return 'std::vector<%s>' % str(self.etype)
    
    def __str__(self):
        return 'std::vector<%s>' % str(self.etype)


class TupleType(MyType):
    def __init__(self, types):
        self.etypes = types

    def set_init_size(self, n):
        self.initsize = n

    def get_init_size(self):
        return self.initsize


class DictType(IterableType):
    def __init__(self, key_ty, value_ty):
        self.key = key_ty
        self.value = value_ty

    def get_elt_type(self):
        return self.key

    def equals(self, ty):
        if not isinstance(ty, DictType):
            return False
        else:
            return ty.key == self.key and ty.value == self.value

    def get_cpp_name(self):
        return 'std::unordered_map<%s, %s>' % (self.key, self.value)

    def __str__(self):
        #return 'std::map<%s, %s>' % (self.key, self.value)
        return 'std::unordered_map<%s, %s>' % (self.key, self.value)
        
class BoolType(MyType):
    def __init__(self):
        self.by_ref = False

    def get_cpp_name(self):
        return 'bool'

    def __str__(self):
        return 'bool'
        
class IntType(MyType):
    def __init__(self, w=64):
        self.by_ref = False
        self.width = w

    def get_width(self):
        return self.width

    def get_cpp_name(self):
        if self.width == 32:
            return 'int'
        elif self.width == 64:
            return 'int64_t'
        else:
            raise Exception()

    def __str__(self):
        if self.width == 32:
            return 'int'
        elif self.width == 64:
            return 'int64_t'
        else:
            raise Exception()

class FloatType(MyType):
    def __init__(self, w=64):
        self.by_ref = False
        self.width = w

    def get_width(self):
        return self.width

    def get_cpp_name(self):
        if self.width == 64:
            return 'double'
        elif self.width == 32:
            return 'float'
        else:
            assert False

    def __str__(self):
        if self.width == 64:
            return 'double'
        elif self.width == 32:
            return 'float'
        else:
            assert False

            
class StrType(MyType):
    def __str__(self):
        return "std::string"

    def get_cpp_name(self):
        return "std::string"
        
    def get_elt_type(self):
        return IntType()


class VoidType(MyType):
    def __init__(self):
        self.by_ref = False
        pass

    def get_cpp_name(self):
        return 'void'

    def __str__(self):
        return 'void'

    
class NpArray(IterableType):
    def __init__(self, dtype, ndim=-1, layout='K'):
        self.etype = dtype
        self.ndim = ndim
        self.layout = layout

    def get_pi_type(self):
        return PINpArray(self.etype, self.ndim, self.layout)

    def get_dimension(self):
        return self.ndim

    def equals(self, a):        
        #return self.etype == a.etype and self.layout == a.layout
        return hasattr(a, 'etype') and self.etype == a.etype

    def __str__(self):
        return "py::array_t<%s>" % self.etype
        #return "pydd::NpArray<%s>*" % self.etype
        
    def get_cpp_name(self):
        return "py::array_t<%s>" % self.etype.get_cpp_name()

    def get_new_str(self):
        return "py::array_t<%s>" % self.etype
        #return "new pydd::NpArray<%s>" % self.etype    

class PINpArray(MyType):
    def __init__(self, dtype, ndim, layout='K'):
        self.etype = dtype
        self.ndim = ndim
        self.layout = layout

    def get_cpp_name(self):
        return "py::array_t<%s>" % self.etype.get_cpp_name()

    def __str__(self):
        return 'py::array_t<%s>' % self.etype

class SparseMat(IterableType):
    def __init__(self, dtype, ndim=2, layout='K'):
        self.etype = dtype
        self.ndim = ndim
        self.layout = layout

    def get_cpp_name(self):
        return 'pydd::Sparse_matrix<%s>' % self.etype.get_cpp_name()    

    def __str__(self):
        return 'pydd::Sparse_matrix<%s>' % self.etype

class Heap(IterableType):
    def __init__(self, key_ty, val_ty):
        self.key = key_ty
        self.val = val_ty
        # self.etype = dtype
    
    def get_elt_type(self):
        return self.key

    def get_cpp_name(self):
        return 'pydd::Heap<%s, %s>' % \
               (self.key.get_cpp_name(), self.val.get_cpp_name())    

    def __str__(self):
        return 'pydd::Heap<%s, %s>' % (self.key, self.val)
    
def pass_by_ref(ty):
    if isinstance(ty, IntType) or \
      isinstance(ty, FloatType) or \
      isinstance(ty, NpArray) or \
      isinstance(ty, SparseMat) or \
      isinstance(ty, Heap) or \
      isinstance(ty, VoidType) or \
      isinstance(ty, StrType) or \
      isinstance(ty, BoolType):
        return False
    else:
        return True

def is_float(ty):
    return isinstance(ty, FloatType)
    
def is_all_float(t1, t2):
    return is_float(t1) and is_float(t2)
    
def is_int(ty):
    return isinstance(ty, IntType)
    
def is_all_int(t1, t2):
    return is_int(t1) and is_int(t2)

def is_array(ty):
    #return isinstance(ty, ListType) or isinstance(ty, NpArray)
    return isinstance(ty, NpArray)

def is_list(ty):
    return isinstance(ty, ListType)

def is_tuple(ty):
    return isinstance(ty, TupleType)


def is_basic_type(ty):
    return not hasattr(ty, 'etype')

    
bool = BoolType()  
int32 = IntType(32)
int64 = IntType(64)
float32 = FloatType(32)
double = float64 = FloatType(64)
float32_list = ListType(float32)
float64_list = ListType(float64)
int32_list = ListType(int32)
int64_list = ListType(int64)
void = VoidType()
string = StrType()
float64_ndarray = NpArray(float64)
float64_1darray = NpArray(float64, 1)
float64_2darray = NpArray(float64, 2)
float32_ndarray = NpArray(float32)
int32_ndarray = NpArray(int32)
int32_1darray = NpArray(int32, 1)
int64_ndarray = NpArray(int64)
bool_ndarray = NpArray(bool)
float64_sparray = SparseMat(float64)
float32_heap = Heap(float32, int32)
int32_int32_heap = Heap(int32, int32)

# Special types for return vars
arg_dependent_type = None



import numpy as np
import intrepydd
from intrepydd.lang import *


def foo(a: Array(float64, 2), r: Array(float64, 2), b: Array(float64, 2), iters: int32):
    for i in range(iters):
        c = 1.0 / (a @ b)
        b = r @ c
    return b
        

foo1 = intrepydd.compile(foo)

N = 100000
a = np.random.randn(N)
b = np.random.randn(N)
assert np.allclose(foo(a, b), foo1(a, b))


from . import defuse
import ast
from . import mytypes
from .utils import get_operator_expansion_func
from . import glb

### Options ###
# 0: Nothing, 1: Debug
verbose_level = 1

# Class used to expand operators into vector library calls
class OperatorExpansion(ast.NodeTransformer):

    def visit_BinOp(self, N: ast.BinOp):
        left = N.left
        right = N.right

        left = self.visit(N.left)
        right = self.visit(N.right)

        # If the expression contains a vector, try to expand the operator into a function
        if hasattr(N, 'type') and mytypes.is_array(N.type):
            func_name = get_operator_expansion_func(N.op)

            # Throw an error when the operator cannot be expanded into a library function
            if func_name == 'n/a':
                msg = "Unidentified operator %s\n" % N.op
                glb.exit_on_node(N, msg)

            if verbose_level >= 1:
                print('[Verbose] found BinOp (%s) to convert into global func: %s' % (N.op, func_name))

            # Transform the BinOp into the correct function call
            func = ast.Name(id=func_name, ctx=ast.Load())
            args = [left, right]

            N = ast.Call(func=func, args=args, keywords=[])

        return N


    def visit_AugAssign(self, N):        
        if hasattr(N.target, "type") and mytypes.is_array(N.target.type):
            op = N.op
            func = ''
            if isinstance(op, ast.Add):
                func = 'plus_eq'   # maybe this should be add_eq
            elif isinstance(op, ast.Sub):
                func = 'sub_eq'
            elif isinstance(op, ast.Mul):
                func = 'mul_eq'
            elif isinstance(op, ast.Div):
                func = 'div_eq'
            else:
                return N

            args = [N.target, N.value]
            call = ast.Call(func=ast.Name(id=func, ctx=ast.Load()), args=args, keywords=[])

            return call
        
        return N




# import subprocess
import sys
import os

# subprocess.call(["ls", "/404"],stderr=sys.stdout.buffer)
import cppimport
sys.path.append(os.getcwd())

module = sys.argv[1]
#print(module)
hashfile = ".%s.cpp.cppimporthash" % module
if os.path.isfile(hashfile):
    os.remove(hashfile)
cppimport.imp(module)




import tracemalloc
import numpy as np
from memory_profiler import memory_usage
from memory_profiler import profile
import gc

#@profile
def foo(a, r, b, iters):
    for i in range(iters):
        #mem_usage_before = memory_usage()
        c = a @ b
        b = r @ (1.0 / c)
        #mem_usage_after = memory_usage()
        #print(f"Memory usage before: {mem_usage_before}, after: {mem_usage_after}")
    return b

#@profile
def foo1(a, r, b, iters):
    c = np.empty_like(a)
    t = np.empty_like(c)
    for i in range(iters):
        #mem_usage_before = memory_usage()
        
        # c = a @ b
        np.matmul(a, b, out=c) 
        # b = r @ c
        np.divide(1.0, c, out=c)
        np.matmul(r, c, out=b)
        
        #mem_usage_after = memory_usage()

        #print(f"Memory usage before: {mem_usage_before}, after: {mem_usage_after}")
    return b


# Example usage with dummy arrays
a = np.random.rand(2000, 2000)
r = np.random.rand(2000, 2000)
b = np.random.rand(2000, 2000)
iters = 3

assert np.allclose(foo(a, r, b, iters), foo1(a, r, b, iters))

#foo(a, r, b, iters)
#foo1(a, r, b, iters)
import timeit
print(timeit.timeit(lambda: foo(a, r, b, iters), number=30) / 30)
print(timeit.timeit(lambda: foo1(a, r, b, iters), number=30) / 30)



import warnings
import logging
import ast
from . import mytypes
from . import glb

#from glb import dump

def deprecated(message):
  def deprecated_decorator(func):
      def deprecated_func(*args, **kwargs):
          warnings.warn("{} is a deprecated function. {}".format(func.__name__, message),
                        category=DeprecationWarning,
                        stacklevel=2)
          warnings.simplefilter('default', DeprecationWarning)
          return func(*args, **kwargs)
      return deprecated_func
  return deprecated_decorator


def get_logger(name):
    return logging.getLogger(name)
  

def setup_logger(logger_name, log_file=None, level=logging.INFO):
    l = logging.getLogger(logger_name)
    formatter = logging.Formatter('%(asctime)s [%(name)s] %(message)s')
    #formatter = logging.Formatter('%(message)s')
    if log_file:
        fileHandler = logging.FileHandler(log_file, mode='w')
        fileHandler.setFormatter(formatter)
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(formatter)

    l.setLevel(level)
    if log_file:
        l.addHandler(fileHandler)
    l.addHandler(streamHandler)
    return l

    
def get_annotation_type(N):
    if isinstance(N, ast.Name):
        id = N.id
        if id == 'int' or id == 'int32' or id == 'I':
            return mytypes.int32
        elif id == 'int64' or id == 'long' or id == 'J':
            return mytypes.int64
        elif id == 'double' or id == 'float64' or id == 'D':
            return mytypes.float64
        elif id == 'float' or id == 'float32' or id == 'F':
            return mytypes.float32
        elif id == 'bool' or id == 'Z':
            return mytypes.bool
    
    elif isinstance(N, ast.Call):
        # Numpy array type
        if N.func.id == 'Array':
            args = N.args
            etype = mytypes.float64                
            ndim = -1
            layout = 'K'
            if len(args) > 0:
                etype = get_annotation_type(args[0])
            if len(args) > 1:
                ndim = args[1].n                
            if len(args) > 2:
                layout = args[2].id
            return mytypes.NpArray(etype, ndim, layout)
        elif N.func.id == 'SparseMat':
            args = N.args
            etype = mytypes.float64
            if len(args) > 0:
                etype = get_annotation_type(args[0])
            return mytypes.SparseMat(etype)
        elif N.func.id == 'Heap':
            args = N.args
            etype = mytypes.float64
            if len(args) > 0:
                etype = get_annotation_type(args[0])
            return mytypes.Heap(etype)
        elif N.func.id == 'Dict':
            args = N.args
            assert len(args) == 2, "Type annotation `Dict` must have two arguments"
            return mytypes.DictType(get_annotation_type(N.args[0]), get_annotation_type(N.args[1]))
        elif N.func.id == 'List':
            return mytypes.ListType(get_annotation_type(N.args[0]))
            
    elif isinstance(N, ast.Subscript):
        if isinstance(N.value, ast.Name):
            if N.value.id == 'List':
                if N.slice.id == 'float':
                    return mytypes.float64_list
                if N.slice.id == 'int':
                    return mytypes.int64_list   
                
    else:
        glb.exit_on_unsupported_type_ann(N)

def get_operator_expansion_func(op):
    if isinstance(op, ast.Add):
        return 'add'
    elif isinstance(op, ast.Div):
        return 'div'
    elif isinstance(op, ast.Eq):
        return 'eq'
    elif isinstance(op, ast.Gt):
        return 'gt'
    elif isinstance(op, ast.GtE):
        return 'ge'
    elif isinstance(op, ast.LtE):
        return 'le'
    elif isinstance(op, ast.Lt):
        return 'lt'
    elif isinstance(op, ast.Mult):
        return 'mul'
    elif isinstance(op, ast.NotEq):
        return 'neq'
    elif isinstance(op, ast.Sub):
        return 'sub'
    else:
        return 'n/a'




from enum import Enum
import ast
from . import glb
from . import libfuncs
from . import mytypes

'''
Sparse and dense array optimizations.

Todo:
 - For sparse, consider 1-D dense matrix with broadcast
  -- sparse_dim: Verify_expression and scalarize_array_op
  -- Valid.MATMULT: Handle matmat_multiply
  -- Valid.TRANSPOSE: Handle innerprod, vectmat, matvect
 - Type system support
  -- dense_opt_recurse: Determine type in addition to current dim support
   --- Use found type for pre_stmts, e.g., pre_stmts[expr] = FOUND_TYPE
  -- eval_node_dense: Use above type for newly created tar_var
 - Standardize options, print (verbose and warning), and assert
'''

### Options ###
# 0: Nothing, 1: Warning, 2: Debug (light), 3: Debug (moderate), 4: Debug (heavy)
verbose_level = 1
dumpfunc = ast.dump
sparse_par = False
dense_par = False
rt_shape_check = True
ew_prestmt = False
test_pldi_algo = False

### Globals ###
ew_funcs = {
    # Map to UnaryOp
    'minus': ast.USub, 'not': ast.Not,
    # Map to BinOp
    'add': ast.Add, 'sub': ast.Sub, 'mul': ast.Mult, 'div': ast.Div,
    'plus_eq': ast.Add, 'minus_eq': ast.Sub,
    # Map to Compare
    'eq': ast.Eq, 'neq': ast.NotEq, 'lt': ast.Lt, 'gt': ast.Gt, 'le': ast.LtE, 'ge': ast.GtE,
    # Map to BoolOp
    'logical_and': ast.And, 'logical_or': ast.Or,
    # Map to no-op
    'set_row': None,
    # Kept as Call (with scalarization)
    'abs': 1, 'isnan': 1, 'isinf': 1, 'sqrt': 1, 'exp': 1,
    'cos': 1, 'sin': 1, 'tan': 1, 'acos': 1, 'asin': 1, 'atan': 1,
    'pow': 2, 'log': 2, 'maximum': 2, 'logical_xor': 2,
}

sparse_dim = 2
set_row_val_idx = 2

### Utilities ###
new_var_count = 0
new_var_base = '__var'
def new_var(ty: mytypes.MyType, func: ast.FunctionDef):
    found = True
    while found:
        global new_var_count, new_var_base
        new_var_count += 1
        name = new_var_base + str(new_var_count)
        found = name in func.typemap
    func.typemap[name] = ty
    return name

def new_name(var: str, ctx: ast.expr_context, ty1 = None, ty2 = None):
    ret = ast.Name(var, ctx())
    if ty1: ret.type = ty1
    elif ty2: ret.type = ty2
    else: ret.type = None
    return ret
    
def new_assign(var: str, expr: ast.expr, ty = None):
    tar = new_name(var, ast.Store, ty, (expr.type if hasattr(expr, 'type') else None))
    ret = ast.Assign(targets = [tar], value = expr, lineno = None, col_offset = None)
    return ret

def new_aug_assign(var: str, op:ast.operator, expr: ast.expr, ty = None):
    tar = new_name(var, ast.Store, ty, (expr.type if hasattr(expr, 'type') else None))
    ret = ast.AugAssign(target = tar, op = op, value = expr)
    return ret

def new_subscript_elem(index):
    if isinstance(index, str):
        return ast.Name(index, ast.Load())
    else:
        return index

def new_subscript(array:str, index, ty = None):
    if isinstance(index, list):
        idxArg = ast.Tuple(list(map(lambda x: new_subscript_elem(x), index)), ast.Load())
    else:
        idxArg = new_subscript_elem(index)
    ret = ast.Subscript(ast.Name(array, ast.Load()), ast.Index(idxArg), ast.Load())
    ret.type = ty
    return ret

def new_call(func: str, args: list, ty = None):
    ret = ast.Call(func = ast.Name(func, ast.Load()), args = args, keywords = [])
    ret.type = ty
    return ret

def contains_var(expr: ast.expr, var: str):
    if isinstance(expr, ast.Name) and expr.id is var:
        return True
    for node in ast.iter_child_nodes(expr):
        if contains_var(node, var):
            return True
    return False

def is_assign_stmt(stmt: ast.stmt):
    return (isinstance(stmt, ast.Assign) or isinstance(stmt, ast.AnnAssign)
            or isinstance(stmt, ast.AugAssign)
            or is_call_to(stmt, 'plus_eq') or is_call_to(stmt, 'minus_eq')
            or is_call_to(stmt, 'set_row'))

def is_call_to(N: ast.expr, name):
    if isinstance(N, ast.Call) and isinstance(N.func, ast.Name):
        fid = N.func.id
    elif (isinstance(N, ast.Expr) and isinstance(N.value, ast.Call)
          and isinstance(N.value.func, ast.Name)):
        fid = N.value.func.id
    else:
        return False

    if isinstance(name, list) or isinstance(name, dict) or isinstance(name, set):
        return fid in name
    else:
        return fid == name

def is_var_with_type(N: ast.expr, func: ast.FunctionDef, type_class, ndim = -1):
    if isinstance(N, ast.Name) and N.id in func.typemap:
        ty = func.typemap[N.id]
        # print(ty, 'vs', (N.type if hasattr(N, 'type') else ast.dump(N))) # TEST
        return isinstance(ty, type_class) and (ndim == -1 or ty.ndim == ndim)
    return False

def is_scalar_var(N: ast.expr, func: ast.FunctionDef):
    if isinstance(N, ast.Num):
        return True
    if isinstance(N, ast.Name) and N.id in func.typemap:
        ty = func.typemap[N.id]
        # print(ty, 'vs', (N.type if hasattr(N, 'type') else ast.dump(N))) # TEST
        return (isinstance(ty, mytypes.IntType) or isinstance(ty, mytypes.FloatType)
                or isinstance(ty, mytypes.BoolType))
    if (isinstance(N, ast.Subscript) and isinstance(N.slice, ast.Index)
        and isinstance(N.value, ast.Name) and N.value.id in func.typemap):
        ty = func.typemap[N.value.id]
        if hasattr(ty, 'ndim'):
            ln = len(N.slice.elts) if isinstance(N.slice, ast.Tuple) else 1
            return ty.ndim == ln
    return False

def is_scalar_expr(N: ast.expr, func: ast.FunctionDef):
    # Todo: Handle BoolOp and Compare
    if isinstance(N, ast.UnaryOp):
        return is_scalar_expr(N.operand, func)
    elif isinstance(N, ast.BinOp):
        return is_scalar_expr(N.left, func) and is_scalar_expr(N.right, func)
    else:
        return is_scalar_var(N, func)

def is_array_slice(expr: ast.expr, func: ast.FunctionDef, is_assign = False):
    if is_assign:
        call_check = is_call_to(expr, 'set_row')
        expected = 3 # Todo: Can change due to set_row signature
    else:
        call_check = is_call_to(expr, 'get_row')
        expected = 2 # Todo: Can change due to get_row signature
    if call_check:
        assert len(expr.args) == expected
        if (is_var_with_type(expr.args[0], func, mytypes.NpArray)
            and (is_scalar_var(expr.args[1], func) or is_scalar_expr(expr.args[1], func))):
            return True
    return False

def get_array_slice_target(expr: ast.expr):
    return expr.args[0]

def get_array_slice_index(expr: ast.expr):
    return expr.args[1]

def collect_arrays(expr: ast.expr, func: ast.FunctionDef, arrays = None):
    if arrays is None:
        arrays = []
    if is_var_with_type(expr, func, mytypes.NpArray):
        arrays.append(expr.id)
    if not is_call_to(expr, 'get_row'): # Todo: Handle array slice
        for node in ast.iter_child_nodes(expr):
            collect_arrays(node, func, arrays)
    return arrays

def get_ret_etype(N: ast.expr):
    if not isinstance(N, ast.Call) or not isinstance(N.func, ast.Name):
        return None
    dummy = [mytypes.float64_ndarray, mytypes.float64_ndarray]
    ty = libfuncs.get_type(N.func.id, 'pydd', dummy)
    # print(ty, 'vs', (N.type if hasattr(N, 'type') else ast.dump(N))) # TEST
    assert ty and hasattr(ty, 'etype')
    return ty.etype


############################################################################################

class ArrayOptimization(ast.NodeVisitor):

    class OptInfo:
        def __init__(self, target_call: ast.Call, full_expr: bool,
                     src_var: str, tar_var: str, pre_stmts: dict,
                     src_slice = None, tar_slice = None):
            self.target_call = target_call  # call to outermost function (e.g., spm_mul)
            self.full_expr = full_expr      # if original stmt is "tar_var = target_call"
            self.src_var = src_var          # sparse/dense matrix name to use
            self.tar_var = tar_var          # sparse/dense matrix name to store
            self.pre_stmts = pre_stmts      # key: ast.Call, value: temp var (type/name)
            self.src_slice = src_slice      # Index if src is array slice
            self.tar_slice = tar_slice      # Index if tar is array slice

        def show(self):
            print('[Verbose] dump OptInfo:')
            if self.target_call: print(' - target_call:', dumpfunc(self.target_call))
            print(' - full_expr:', self.full_expr)
            if self.src_var: print(' - src_var:', self.src_var)
            if self.tar_var: print(' - tar_var:', self.tar_var)
            if self.src_slice: print(' - src_slice:', dumpfunc(self.src_slice))
            if self.tar_slice: print(' - tar_slice:', dumpfunc(self.tar_slice))
            for expr in self.pre_stmts: print(' - pre_stmt:', dumpfunc(expr))

    class Valid(Enum):
        INVALID = -1
        SCALAR = 0
        S_MAT = 1
        D_MAT = 2
        S_MUL = 3
        D_ELEM = 4
        MATMULT = 5
        TRANSPOSE = 6

    def __init__(self):
        self.curr_func = None
        self.enable_sparse = False
        self.enable_dense = False

    def verify_expression(self, expr: ast.expr, src_spm: str, pre_stmts: dict, set_pre = True):
        res = self.Valid.INVALID

        if is_scalar_var(expr, self.curr_func):
            res = self.Valid.SCALAR
        elif is_scalar_expr(expr, self.curr_func):
            if set_pre:
                pre_stmts[expr] = expr.type if hasattr(expr, 'type') else mytypes.float64
            res = self.Valid.SCALAR
        elif isinstance(expr, ast.Name) and expr.id == src_spm:
            res = self.Valid.S_MAT
        elif is_var_with_type(expr, self.curr_func, mytypes.NpArray, sparse_dim):
            res = self.Valid.D_MAT
        elif is_call_to(expr, 'transpose'):
            if set_pre:
                pre_stmts[expr] = mytypes.float64_2darray
                res = self.Valid.D_MAT
            else:
                res = self.Valid.TRANSPOSE
        elif is_call_to(expr, 'matmult'):
            v = list(map(lambda x: self.verify_expression(x, src_spm, pre_stmts, False), expr.args))
            if ((v[0] == self.Valid.D_MAT or v[0] == self.Valid.D_ELEM
                 or v[0] == self.Valid.MATMULT or v[0] == self.Valid.TRANSPOSE)
                and (v[1] == self.Valid.D_MAT or v[1] == self.Valid.D_ELEM
                     or v[1] == self.Valid.MATMULT or v[1] == self.Valid.TRANSPOSE)):
                if set_pre:
                    if v[0] != self.Valid.D_MAT: pre_stmts[expr.args[0]] = mytypes.float64_2darray
                    if v[1] != self.Valid.D_MAT: pre_stmts[expr.args[1]] = mytypes.float64_2darray
                res = self.Valid.MATMULT
        elif isinstance(expr, ast.Call):
            v = list(map(lambda x: self.verify_expression(x, src_spm, pre_stmts), expr.args))
            if (is_call_to(expr, 'spm_mul')
                and (((v[0] == self.Valid.S_MAT or v[0] == self.Valid.S_MUL)
                      and v[1] != self.Valid.INVALID)
                     or ((v[1] == self.Valid.S_MAT or v[1] == self.Valid.S_MUL)
                         and v[0] != self.Valid.INVALID))):
                res = self.Valid.S_MUL
            elif is_call_to(expr, ew_funcs):
                if ((len(v) == 1
                     and (v[0] == self.Valid.D_MAT
                          or v[0] == self.Valid.D_ELEM or v[0] == self.Valid.MATMULT))
                    or (len(v) == 2
                        and (v[0] == self.Valid.SCALAR or v[0] == self.Valid.D_MAT
                             or v[0] == self.Valid.D_ELEM or v[0] == self.Valid.MATMULT)
                        and (v[1] == self.Valid.SCALAR or v[1] == self.Valid.D_MAT
                             or v[1] == self.Valid.D_ELEM or v[1] == self.Valid.MATMULT)
                        and (v[0] != self.Valid.SCALAR or v[1] != self.Valid.SCALAR))):
                    res = self.Valid.D_ELEM

        return res

    def contains_func_call(self, expr: ast.expr):
        if isinstance(expr, ast.Call) and isinstance(expr.func, ast.Name):
            return True
        for node in ast.iter_child_nodes(expr):
            if self.contains_func_call(node):
                return True
        return False

    def verify_and_collect_info(self, stmt: ast.stmt, target_call: ast.Call):
        if not any(list(map(lambda x: self.contains_func_call(x), target_call.args))):
            if verbose_level >= 2: print('[Verbose] No function to fuse for spm_mul.')
            return None

        src_spm = None
        arg0 = target_call.args[0]
        while arg0:
            if is_call_to(arg0, 'spm_mul'):
                arg0 = arg0.args[0]
            else:
                if is_var_with_type(arg0, self.curr_func, mytypes.SparseMat):
                    src_spm = arg0.id
                arg0 = None
        if not src_spm:
            if verbose_level >= 2: print('[Verbose] No src_spm found.')
            return None

        pre_stmts = {}
        if self.verify_expression(target_call, src_spm, pre_stmts) != self.Valid.S_MUL:
            if verbose_level >= 2: print('[Verbose] Invalid expression.')
            return None

        if ((isinstance(stmt, ast.Assign) or isinstance(stmt, ast.AnnAssign))
            and target_call is stmt.value and stmt.targets[0].id is not src_spm):
            assert len(stmt.targets) == 1
            assert is_var_with_type(stmt.targets[0], self.curr_func, mytypes.SparseMat)
            tar_spm = stmt.targets[0].id
            full_expr = True
        else:
            ty = self.curr_func.typemap[src_spm]
            tar_spm = new_var(ty, self.curr_func)
            full_expr = False

        opt_info = self.OptInfo(target_call, full_expr, src_spm, tar_spm, pre_stmts)
        if verbose_level >= 2:
            print('[Verbose] spm_mul with elem-wise operations found.')
            if verbose_level >= 3:
                opt_info.show()

        return opt_info

    def sparse_opt_analysis(self, stmt: ast.stmt, expr: ast.expr, opt_list: list):
        if is_call_to(expr, 'spm_mul'):
            opt_info = self.verify_and_collect_info(stmt, expr)
            if opt_info:
                opt_list.append(opt_info)
            else:
                for arg in expr.args:
                    self.sparse_opt_analysis(stmt, arg, opt_list)
        else:
            for node in ast.iter_child_nodes(expr):
                self.sparse_opt_analysis(stmt, node, opt_list)

    def find_reusable_var(self, opt_list: list, opt_info: OptInfo, pre_stmt: ast.expr):
        # Todo: return var name if same expression exists
        return None

    def replace_func_arg_by_var(self, expr, target_expr: ast.expr, var: str):
        if isinstance(expr, ast.stmt) and expr.value is target_expr:
            expr.value = ast.Name(var, ast.Load())
            
        if isinstance(expr, ast.Call):
            for i in range(len(expr.args)):
                if expr.args[i] is target_expr:
                    expr.args[i] = ast.Name(var, ast.Load())

        for node in ast.iter_child_nodes(expr):
            self.replace_func_arg_by_var(node, target_expr, var)

    def assign_new_vars(self, opt_list: list):
        for opt_info in opt_list:
            for expr in opt_info.pre_stmts:
                var = self.find_reusable_var(opt_list, opt_info, expr)
                if not var:
                    ty = opt_info.pre_stmts[expr]
                    var = new_var(ty, self.curr_func)
                    opt_info.pre_stmts[expr] = var
                self.replace_func_arg_by_var(opt_info.target_call, expr, var)

    def update_pre_stmts(self, start: int, opt_list: list, target_expr: ast.expr, var: str):
        for i in range(start, len(opt_list)):
            for expr in opt_list[i].pre_stmts:
                self.replace_func_arg_by_var(expr, target_expr, var)

    def generate_init_spm(self, src_spm: str, tar_spm: str):
        shape0 = new_call('shape', [ast.Name(src_spm, ast.Load()), ast.Num(0)])
        shape1 = new_call('shape', [ast.Name(src_spm, ast.Load()), ast.Num(1)])
        return new_assign(tar_spm, new_call('empty_spm', [shape0, shape1]))

    def generate_for(self, idx: str, func: str, arg0: str, arg1, body: list, is_pfor = False):
        arg1Ast = ast.Num(arg1) if isinstance(arg1, int) else ast.Name(arg1, ast.Load())
        rangeCall = new_call('range', [new_call(func, [ast.Name(arg0, ast.Load()), arg1Ast])])
        ret = ast.For(ast.Name(idx, ast.Store()), rangeCall, body, [], lineno = None, col_offset = None)
        ret.is_pfor = is_pfor
        if is_pfor: self.curr_func.has_pfor = True

        return ret

    def scalarize_matmult(self, args: list, row: str, col: str, stmts: list):
        assert len(args) == 2
        assert is_var_with_type(args[0], self.curr_func, mytypes.NpArray, sparse_dim)
        assert is_var_with_type(args[1], self.curr_func, mytypes.NpArray, sparse_dim)

        res = new_var(mytypes.float64, self.curr_func)
        stmts.append(new_assign(res, ast.Num(0.0)))

        idx = new_var(mytypes.int32, self.curr_func)
        left = new_subscript(args[0].id, [row, idx])
        right = new_subscript(args[1].id, [idx, col])
        aug = new_aug_assign(res, ast.Add(), ast.BinOp(left, ast.Mult(), right))

        for_stmt = self.generate_for(idx, 'shape', args[0].id, 1, [aug])
        stmts.append(for_stmt)
        return ast.Name(res, ast.Load())

    def scalarize_array_op(self, expr: ast.expr, src_spm: str,
                           row: str, col: str, value: str, stmts: list):
        if is_scalar_var(expr, self.curr_func):
            return expr
        elif isinstance(expr, ast.Name) and expr.id == src_spm:
            return ast.Name(value, ast.Load())
        elif is_var_with_type(expr, self.curr_func, mytypes.NpArray, sparse_dim):
            return new_subscript(expr.id, [row, col])
        elif is_call_to(expr, 'matmult'):
            return self.scalarize_matmult(expr.args, row, col, stmts)
        elif is_call_to(expr, 'spm_mul'):
            left = self.scalarize_array_op(expr.args[0], src_spm, row, col, value, stmts)
            right = self.scalarize_array_op(expr.args[1], src_spm, row, col, value, stmts)
            assert left and right
            return ast.BinOp(left, ast.Mult(), right)
        elif is_call_to(expr, ew_funcs):
            ewf = ew_funcs[expr.func.id]
            if isinstance(ewf, int):
                assert len(expr.args) == ewf
                for i in range(ewf):
                    expr.args[i] = self.scalarize_array_op(expr.args[i], src_spm,
                                                           row, col, value, stmts)
                return expr
            else:
                op = ewf()
                if len(expr.args) == 1:
                    operand = self.scalarize_array_op(expr.args[0], src_spm, row, col, value, stmts)
                    assert operand and isinstance(op, ast.unaryop)
                    return ast.UnaryOp(op, operand)
                elif len(expr.args) == 2:
                    left = self.scalarize_array_op(expr.args[0], src_spm, row, col, value, stmts)
                    right = self.scalarize_array_op(expr.args[1], src_spm, row, col, value, stmts)
                    assert left and right
                    if isinstance(op, ast.operator): return ast.BinOp(left, op, right)
                    elif isinstance(op, ast.cmpop): return ast.Compare(left, [op], [right])
                    elif isinstance(op, ast.boolop): return ast.BoolOp(op, [left, right])
        return None

    def generate_body(self, target_call: ast.expr, src_spm: str, tar_spm: str,
                      row: str, col: str, cidx: str, value: str):
        col_def = new_assign(col, new_call('getcol', [ast.Name(src_spm, ast.Load()),
                                                      ast.Name(row, ast.Load()),
                                                      ast.Name(cidx, ast.Load())]))
        val_def = new_assign(value, new_call('getval', [ast.Name(src_spm, ast.Load()),
                                                          ast.Name(row, ast.Load()),
                                                          ast.Name(cidx, ast.Load())]))
        stmts = [col_def, val_def]

        body_expr = self.scalarize_array_op(target_call, src_spm, row, col, value, stmts)
        assert body_expr
        stmts.append(ast.Expr(new_call('spm_set_item_unsafe',
                                       [ast.Name(tar_spm, ast.Load()), body_expr,
                                        ast.Name(row, ast.Load()), ast.Name(col, ast.Load())])))
        return stmts

    def sparse_opt_codegen(self, opt_info: OptInfo):
        stmts = []
        for expr in opt_info.pre_stmts:
            var = opt_info.pre_stmts[expr]
            if var:
                stmts.append(new_assign(var, expr))

        row = new_var(mytypes.int32, self.curr_func)
        cidx = new_var(mytypes.int32, self.curr_func)
        col = new_var(mytypes.int32, self.curr_func)
        value = new_var(mytypes.float64, self.curr_func)

        init = self.generate_init_spm(opt_info.src_var, opt_info.tar_var)
        stmts.append(init)

        body = self.generate_body(opt_info.target_call, opt_info.src_var, opt_info.tar_var,
                                  row, col, cidx, value)
        for_inner = self.generate_for(cidx, 'getnnz', opt_info.src_var, row, body)
        for_outer = self.generate_for(row, 'shape', opt_info.src_var, 0, [for_inner], sparse_par)
        stmts.append(for_outer)
        if verbose_level >= 4:
            print('[Verbose] Generated code:')
            for stmt in stmts: print(dumpfunc(stmt))

        return stmts


    def is_fusable_elemwise(self, expr: ast.expr, dims: list, ew_prs: list):
        global set_row_val_idx
        if is_call_to(expr, 'set_row'):
            return is_array_slice(expr, self.curr_func, True) and dims[set_row_val_idx] != -1

        args = expr.args
        n = len(args)
        assert n == len(dims)

        if min(dims) == -1:
            return False

        for i in range(n):
            if dims[i] < max(dims) and is_call_to(args[i], ew_funcs):
                if not ew_prestmt:
                    return False
                ew_prs.append(i)

        return True

    def dense_opt_recurse(self, stmt: ast.stmt, expr: ast.expr, opt_list: list):
        global set_row_val_idx

        dim = -1
        pre_stmts = {}
        if is_scalar_var(expr, self.curr_func):
            dim = 0
        elif is_scalar_expr(expr, self.curr_func):
            pre_stmts[expr] = expr.type if hasattr(expr, 'type') else mytypes.float64
            dim = 0
        elif is_var_with_type(expr, self.curr_func, mytypes.NpArray):
            dim = self.curr_func.typemap[expr.id].ndim
        elif is_array_slice(expr, self.curr_func):
            tar_dim = self.curr_func.typemap[get_array_slice_target(expr).id].ndim
            if tar_dim > 0:
                dim = tar_dim - 1
        elif isinstance(expr, ast.Call):
            n = len(expr.args)
            dims = list(range(n))
            pres = list(range(n))
            for i in range(n):
                dims[i], pres[i] = self.dense_opt_recurse(stmt, expr.args[i], opt_list)

            ew_prs = []
            if is_call_to(expr, ew_funcs) and self.is_fusable_elemwise(expr, dims, ew_prs):
                s = set_row_val_idx if is_call_to(expr, 'set_row') else 0
                dim = max(dims[s:])
                for i in range(s, n):
                    if i in ew_prs:
                        if not self.eval_node_dense(stmt, expr.args[i], dims[i], pres[i], opt_list):
                            assert dims[i] >= 1 and isinstance(expr, ast.Call)
                            etype = get_ret_etype(expr)
                            pre_stmts[expr.args[i]] = mytypes.NpArray(etype, dims[i])
                    else:
                        pre_stmts.update(pres[i])
            else:
                for i in range(n):
                    self.eval_node_dense(stmt, expr.args[i], dims[i], pres[i], opt_list)

                if is_call_to(expr, 'transpose'):
                    if verbose_level >= 1 and dims[0] != 2:
                        print('[Warning] transpose assumes a 2-D array as argument.')
                    dim = 2
                elif is_call_to(expr, 'spmv'):
                    if verbose_level >= 1 and (dims[0] != 1 or dims[1] != 1):
                        print('[Warning] spmv assumes 2-D/1-D arrays as arguments.')
                    dim = 1
                elif is_call_to(expr, 'spmm_dense'):
                    if verbose_level >= 1 and (dims[0] != 2 or dims[1] != 2):
                        print('[Warning] spmm_dense assumes 2-D arrays as arguments.')
                    dim = 1
                elif is_call_to(expr, 'matmat_multiply'):
                    if verbose_level >= 1 and (dims[0] != 2 or dims[1] != 2):
                        print('[Warning] matmat_multiply assumes 2-D arrays as arguments.')
                    dim = 2
                elif is_call_to(expr, 'vectmat_multiply'):
                    if verbose_level >= 1 and (dims[0] != 1 or dims[1] != 2):
                        print('[Warning] vectmat_multiply assumes 1-D/2-D arrays as arguments.')
                    dim = 1
                elif is_call_to(expr, 'matvect_multiply'):
                    if verbose_level >= 1 and (dims[0] != 2 or dims[1] != 1):
                        print('[Warning] matvect_multiply assumes 2-D/1-D arrays as arguments.')
                    dim = 1
                elif is_call_to(expr, 'innerprod'):
                    if verbose_level >= 1 and (dims[0] != 1 or dims[1] != 1):
                        print('[Warning] innerprod assumes 1-D arrays as arguments.')
                    dim = 0
                elif is_call_to(expr, 'matmult') and dims[0] >= 1 and dims[1] >= 1:
                    assert dims[0] <= 2 and dims[1] <= 2
                    dim = dims[0] + dims[1] - 2
                    # 2+2: matmat, 1+2: vectmat, 2+1: matvect, 1+1: innerprod

                if dim == 0:
                    pre_stmts[expr] = mytypes.float64
                elif dim >= 1:
                    pre_stmts[expr] = mytypes.NpArray(mytypes.float64, dim)

        if test_pldi_algo: expr.ndims = dim  # Test for PLDI algo
        return dim, pre_stmts

    def get_src_var_without_slice(self, expr: ast.expr, pre_stmts: dict, type_class, ndim = -1):
        if isinstance(expr, ast.Name) and expr.id in self.curr_func.typemap:
            ty = self.curr_func.typemap[expr.id]
            if isinstance(ty, type_class) and (ndim == -1 or ty.ndim == ndim):
                return expr.id

        if not is_array_slice(expr, self.curr_func):
            for node in ast.iter_child_nodes(expr):
                if not node in pre_stmts:
                    var = self.get_src_var_without_slice(node, pre_stmts, type_class, ndim)
                    if var:
                        return var
        return None

    def get_src_var_with_slice(self, expr: ast.expr, pre_stmts: dict, type_class, ndim = -1):
        if is_array_slice(expr, self.curr_func):
            arr = get_array_slice_target(expr)
            if isinstance(arr, ast.Name) and arr.id in self.curr_func.typemap:
                ty = self.curr_func.typemap[arr.id]
                if isinstance(ty, type_class) and (ndim == -1 or ty.ndim - 1 == ndim):
                    return arr.id, get_array_slice_index(expr)

        for node in ast.iter_child_nodes(expr):
            if not node in pre_stmts:
                var, slice = self.get_src_var_with_slice(node, pre_stmts, type_class, ndim)
                if var and slice:
                    return var, slice
        return None, None

    def get_src_var(self, expr: ast.expr, pre_stmts: dict, type_class, ndim = -1):
        src_var = self.get_src_var_without_slice(expr, pre_stmts, type_class, ndim)
        if src_var:
            return src_var, None

        return self.get_src_var_with_slice(expr, pre_stmts, type_class, ndim)

    def is_nesting_fusable_funcs(self, args: list):
        for arg in args:
            if is_call_to(arg, ew_funcs) or is_array_slice(arg, self.curr_func):
                return True
        return False

    def get_target_var(self, stmt: ast.stmt, expr: ast.expr):
        tar_var, tar_slice = None, None
        if (isinstance(stmt, ast.Assign) or isinstance(stmt, ast.AnnAssign)) and expr is stmt.value:
            assert len(stmt.targets) == 1
            assert is_var_with_type(stmt.targets[0], self.curr_func, mytypes.NpArray)
            tar_var = stmt.targets[0].id
        elif is_call_to(stmt, 'plus_eq') or is_call_to(stmt, 'minus_eq'):
            arg = stmt.value.args[0]
            if is_array_slice(arg, self.curr_func):
                tar_var = get_array_slice_target(arg).id
                tar_slice = get_array_slice_index(arg)
            else:
                assert is_var_with_type(arg, self.curr_func, mytypes.NpArray)
                tar_var = arg.id
        elif is_call_to(stmt, 'set_row'):
            assert is_array_slice(expr, self.curr_func, True)
            tar_var = get_array_slice_target(expr).id
            tar_slice = get_array_slice_index(expr)

        return tar_var, tar_slice

    def eval_node_dense(self, stmt: ast.stmt, expr: ast.expr,
                        dim: int, pre_stmts: dict, opt_list: list):
        global set_row_val_idx

        if dim <= 0 or not is_call_to(expr, ew_funcs):
            return None
        elif not self.is_nesting_fusable_funcs(expr.args):
            if verbose_level >= 2: print('[Verbose] No nested dense elem-wise operations.')
            return None
        elif verbose_level >= 2: print('[Verbose] Nested dense elem-wise operations found.')

        if is_call_to(expr, 'set_row'):
            target_call = expr.args[set_row_val_idx]
        else:
            target_call = expr
        assert isinstance(target_call, ast.Call)

        src_var, src_slice = self.get_src_var(target_call, pre_stmts, mytypes.NpArray, dim)
        assert src_var

        tar_var, tar_slice = self.get_target_var(stmt, expr)
        if tar_var:
            ty0 = self.curr_func.typemap[tar_var]
            if ty0.ndim == -1:
                self.curr_func.typemap[tar_var] = mytypes.NpArray(ty0.etype, dim)
            else:
                ty0_ndim = ty0.ndim - 1 if tar_slice else ty0.ndim
                assert ty0_ndim == dim
            full_expr = True
        else:
            ty = mytypes.NpArray(get_ret_etype(target_call), dim)
            tar_var = new_var(ty, self.curr_func)
            full_expr = False

        opt_info = self.OptInfo(target_call, full_expr, src_var, tar_var, pre_stmts,
                                src_slice, tar_slice)
        opt_list.append(opt_info)
        if verbose_level >= 3:
            opt_info.show()

        return opt_info

    # Test for PLDI algo
    def to_split(self, N: ast.Call, A: ast.Call):
        ewf1 = is_call_to(N, ew_funcs) and N.ndims >= 1
        ewf2 = is_call_to(A, ew_funcs) and A.ndims >= 1
        return ewf1 != ewf2 or ewf1 and ewf2 and N.ndims != A.ndims

    def split_subtrees_for_fusion(self, N, trees: dict):
        if isinstance(N, ast.Assign) or isinstance(N, ast.AnnAssign):
            trees[N.targets[0].id] = N.value
            self.split_subtrees_for_fusion(N.value, trees)
        elif isinstance(N, ast.Call):
            for A in reversed(N.args):
                if isinstance(A, ast.Call):
                    if self.to_split(N, A):
                        v = new_var(mytypes.NpArray(mytypes.float64, A.ndims), self.curr_func)
                        trees[v] = A
                    self.split_subtrees_for_fusion(A, trees)

    def test_PLDI_dense_algo(self, stmt: ast.stmt):
        trees = {}
        self.split_subtrees_for_fusion(stmt, trees)
        print('Test for PLDI dense algo:')
        for v in trees:
            print(v+':', dumpfunc(trees[v]))

    def dense_opt_analysis(self, stmt: ast.stmt, expr: ast.expr, opt_list: list):
        dim, pre_stmts = self.dense_opt_recurse(stmt, expr, opt_list)
        self.eval_node_dense(stmt, expr, dim, pre_stmts, opt_list)

        if test_pldi_algo: self.test_PLDI_dense_algo(stmt)  # Test for PLDI algo

    def generate_compatibility_check(self, target_call: ast.Call):
        arrays = collect_arrays(target_call, self.curr_func)
        if len(arrays) >= 2:
            args0 = list(map(lambda x: ast.Name(x, ast.Load()), arrays))
            check = new_call('compatibility_check', args0)
            # args = ast.List(args0, ast.Load())
            # args.type = mytypes.ListType(mytypes.float64_ndarray)
            # check = new_call('compatibility_check', [args])
            return ast.Expr(check)
        return None

    def generate_init_dense(self, src_var: str, tar_var: str, iv: list):
        ss = list(map(lambda x: new_call('shape', [ast.Name(src_var, ast.Load()), ast.Num(x)]), iv))
        shapes = ast.List(ss, ast.Load())
        shapes.type = mytypes.int64_list

        et = self.curr_func.typemap[tar_var].etype
        if isinstance(et, mytypes.IntType):
            ef = 'int32' if et.width == 32 else 'int64'
        elif isinstance(et, mytypes.FloatType):
            ef = 'float32' if et.width == 32 else 'float64'
        elif isinstance(et, mytypes.BoolType):
            ef = 'boolean'
        else:
            assert False
        etype = new_call(ef, [])
        return new_assign(tar_var, new_call('empty', [shapes, etype]))

    def scalarize_dense_array_op(self, expr: ast.expr, indices: list):
        if is_scalar_var(expr, self.curr_func):
            return expr
        elif is_var_with_type(expr, self.curr_func, mytypes.NpArray):
            n = len(indices) - self.curr_func.typemap[expr.id].ndim
            return new_subscript(expr.id, indices[n:])
        elif is_array_slice(expr, self.curr_func):
            arr = get_array_slice_target(expr)
            idx = get_array_slice_index(expr)
            n = len(indices) - (self.curr_func.typemap[arr.id].ndim - 1)
            return new_subscript(arr.id, [idx] + indices[n:])
        elif is_call_to(expr, ew_funcs):
            ewf = ew_funcs[expr.func.id]
            if isinstance(ewf, int):
                nargs = len(expr.args)
                if verbose_level >= 1 and nargs != ewf:
                    print('[Waraning] Unmatched # args in for func:', expr.func.id)
                for i in range(nargs):
                    expr.args[i] = self.scalarize_dense_array_op(expr.args[i], indices)
                return expr
            else:
                op = ewf()
                if len(expr.args) == 1:
                    operand = self.scalarize_dense_array_op(expr.args[0], indices)
                    assert operand and isinstance(op, ast.unaryop)
                    return ast.UnaryOp(op, operand)
                elif len(expr.args) == 2:
                    left = self.scalarize_dense_array_op(expr.args[0], indices)
                    right = self.scalarize_dense_array_op(expr.args[1], indices)
                    assert left and right
                if isinstance(op, ast.operator): return ast.BinOp(left, op, right)
                elif isinstance(op, ast.cmpop): return ast.Compare(left, [op], [right])
                elif isinstance(op, ast.boolop): return ast.BoolOp(op, [left, right])
        return None

    def dense_opt_codegen(self, opt_info: OptInfo):
        stmts = []
        for expr in opt_info.pre_stmts:
            var = opt_info.pre_stmts[expr]
            if var:
                stmts.append(new_assign(var, expr))

        src_dim = self.curr_func.typemap[opt_info.src_var].ndim
        src_lw = 1 if opt_info.src_slice else 0
        idxvec = list(range(src_lw, src_dim))
        indices = list(map(lambda x: new_var(mytypes.int32, self.curr_func), idxvec))

        if rt_shape_check:
            check = self.generate_compatibility_check(opt_info.target_call)
            if check:
                stmts.append(check)

        if (not opt_info.full_expr
            or (not opt_info.tar_slice
                and not contains_var(opt_info.target_call, opt_info.tar_var))):
            init = self.generate_init_dense(opt_info.src_var, opt_info.tar_var, idxvec)
            stmts.append(init)

        body_expr = self.scalarize_dense_array_op(opt_info.target_call, indices)
        assert body_expr
        idcs = [opt_info.tar_slice] + indices if opt_info.tar_slice else indices
        loop = ast.Assign([new_subscript(opt_info.tar_var, idcs)], body_expr, lineno=None, col_offset=None)
        for i in reversed(idxvec):
            pf = dense_par and i == src_lw
            loop = self.generate_for(indices[i-src_lw], 'shape', opt_info.src_var, i, [loop], pf)
        assert isinstance(loop, ast.For)
        stmts.append(loop)
        if verbose_level >= 4:
            print('[Verbose] Generated code:')
            for stmt in stmts: print(dumpfunc(stmt))

        return stmts


    def array_opt(self, stmts: list, analysis, codegen):
        new_stmts = []
        transformed = False
        for stmt in stmts:
            if is_assign_stmt(stmt):
                opt_list = []
                analysis(stmt, stmt.value, opt_list)
                if opt_list:
                    transformed = True
                    self.assign_new_vars(opt_list)
                    for opt_info in opt_list:
                        new_stmts.extend(codegen(opt_info))
                        if not opt_info.full_expr:
                            self.replace_func_arg_by_var(stmt,
                                                         opt_info.target_call,
                                                         opt_info.tar_var)
                            self.update_pre_stmts(opt_list.index(opt_info)+1, opt_list,
                                                  opt_info.target_call,
                                                  opt_info.tar_var)

                    if not opt_list[-1].full_expr: # Only Last can be full_expr
                        if verbose_level >= 4:
                            print('[Verbose] Replaced statement:\n', dumpfunc(stmt))
                        new_stmts.append(stmt)
                else:
                    new_stmts.append(stmt)
            else:
                self.visit(stmt)
                new_stmts.append(stmt)

        if transformed:
            stmts.clear()
            stmts.extend(new_stmts)

    def visit_body(self, stmts: list):
        if self.enable_sparse:
            self.array_opt(stmts, self.sparse_opt_analysis, self.sparse_opt_codegen)
        if self.enable_dense:
            self.array_opt(stmts, self.dense_opt_analysis, self.dense_opt_codegen)


############################################################################################

    def visit_FunctionDef(self, N: ast.FunctionDef):
        if verbose_level >= 2: print('[Verbose] Array opt for function:', N.name)
        self.curr_func = N
        self.enable_sparse = 'sparse_opt' in glb.options and glb.options['sparse_opt']
        self.enable_dense = 'dense_opt' in glb.options and glb.options['dense_opt']
        self.visit_body(N.body)

    def visit_For(self, N: ast.For):
        self.visit_body(N.body)
        assert not N.orelse

    def visit_While(self, N: ast.While):
        self.visit_body(N.body)
        assert not N.orelse

    def visit_If(self, N: ast.If):
        self.visit_body(N.body)
        self.visit_body(N.orelse)



import timeit
import numpy as np
import numba
import intrepydd
from test_array_add_mul import foo


N = 1000000
a = np.random.randn(N)
b = np.random.randn(N)
print('np:', timeit.timeit(lambda: foo(a, b), number=1000) / 1000)

foo_pydd = intrepydd.compile(foo)
foo_nb = numba.jit(foo)
print('pydd:', timeit.timeit(lambda: foo_pydd(a, b), number=1000) / 1000)
print('numba:', timeit.timeit(lambda: foo_nb(a, b), number=1000) / 1000)


import ast
from . import libfuncs
from . import glb
'''
Support AST transformations that are independent from type information.

Todo:
 - Standardize options, print (verbose and warning), and assert
'''

### Options ###
# 0: Nothing, 1: Debug
verbose_level = 0

class TypeFreeTransformer(ast.NodeTransformer):
    def visit_Call(self, N):
        '''
        Conversion: "A.abs()" into "abs(A)"
        '''
        candidate = isinstance(N.func, ast.Attribute) and libfuncs.is_global_func(N.func.attr)
        if candidate and isinstance(N.func.value, ast.Name):
            candidate = not glb.cpp_module.is_module(N.func.value.id)

        if candidate:
            if verbose_level >= 1:
                print('[Verbose] found Attribute to convert into global func:', N.func.attr)

            
            func_name = N.func.attr
            func_ctx = self.visit(N.func.ctx)
            func_arg = self.visit(N.func.value)
            N.func = ast.Name(id=func_name, ctx=func_ctx)
            args = [func_arg]
        else:
            N.func = self.visit(N.func)
            args = []

        for arg in N.args:
            args.append(self.visit(arg))
        N.args = args
        keys = []
        for key in N.keywords:
            keys.append(self.visit(key))
        N.keys = keys

        return N

    def visit_Attribute(self, N):
        if N.attr == 'T':
            if verbose_level >= 1:
                print('[Verbose] found Attribute to convert into global func: transpose')

            # Conversion: "A.T" into "transpose(A)"
            func = ast.Name(id='transpose', ctx=ast.Load())
            args = [self.visit(N.value)]
            N = ast.Call(func=func, args=args, keywords=[])
        else:
            #print(ast.dump(N))
            N.value = self.visit(N.value)
            N.ctx = self.visit(N.ctx)

        return N

    def visit_BinOp(self, N):
        if isinstance(N.op, ast.MatMult):
            if verbose_level >= 1:
                print('[Verbose] found BinOp (@) to convert into global func: matmult')

            # Conversion: "A @ B" into "matmult(A, B)"
            func = ast.Name(id='matmult', ctx=ast.Load())
            arg1 = self.visit(N.left)
            arg2 = self.visit(N.right)
            args = [arg1, arg2]
            N = ast.Call(func=func, args=args, keywords=[])
        else:
            N.left = self.visit(N.left)
            N.op = self.visit(N.op)
            N.right = self.visit(N.right)

        return N





import ast
from . import mytypes
from . import glb
#from glb import dump


def get_line_number(N):
    return N.lineno


def get_lineno(N):
    return N.lineno


def has_lineno(N):
    return hasattr(N, 'lineno')


def has_location(N):
    return hasattr(N, 'lineno') and hasattr(N, 'col_offset')
    
def is_name(N):
    return isinstance(N, ast.Name)


def is_assign(N):
    return isinstance(N, ast.Assign) or isinstance(N, ast.AnnAssign)


def is_aug_assign(N):
    return isinstance(N, ast.AugAssign)


def is_assign_to_name(N):
    if is_assign(N):
        targets = N.targets
        assert len(targets) == 1, "multiple targets in Assign not supported"
        return is_name(targets[0])
    return False


def is_alloc_site(N):
    if not is_assign(N):
        return False
    rhs = N.value

    # if 'zeros' in ast.dump(rhs):
    #     print('call', rhs)
    # if is_call(rhs):
    #     print('call', rhs)
    return is_call_to(rhs, 'empty') or \
        is_call_to(rhs, 'zeros') or \
        is_call_to(rhs, 'empty_like') or \
        is_call_to(rhs, 'zeros_like')


def is_bin_op(N):
    return isinstance(N, ast.BinOp)


def get_bin_op_left(N):
    return N.left


def get_bin_op_right(N):
    return N.right


def get_bin_op_operands(N):
    return (N.left, N.right)


def is_setitem(N):
    if is_assign(N):
        tar = N.targets[0]
        return isinstance(tar, ast.Subscript)
    return False


def get_setitem_base(N: ast.Assign):
    return N.targets[0].value.id


def get_setitem_value(N: ast.Assign):
    return N.value


def is_call_to(N: ast.expr, name):
    if isinstance(N, ast.Call) and isinstance(N.func, ast.Name):
        fid = N.func.id
    elif (isinstance(N, ast.Expr) and isinstance(N.value, ast.Call)
          and isinstance(N.value.func, ast.Name)):
        fid = N.value.func.id
    else:
        return False

    if isinstance(name, list) or isinstance(name, dict) or isinstance(name, set):
        return fid in name
    else:
        return fid == name


def is_call(N):
    return isinstance(N, ast.Call)


def create_call(funcname, argnames):
    if type(argnames) == str:
        argnames = [argnames]
    args = []
    for arg in argnames:
        args.append(ast.Name(id=arg, ctx=ast.Load()))
    call = ast.Call(func=ast.Name(id=funcname, ctx=ast.Load()), args=args, keywords=[])
    
    return call
    

def is_local_def(N):
    return is_assign_to_name(N)


def get_defined_name(N) -> str:
    return N.targets[0].id


def is_for(N):
    return isinstance(N, ast.For)


def is_while(N):
    return isinstance(N, ast.While)


def is_loop(N):
    return isinstance(N, ast.For) or isinstance(N, ast.While)


def get_loop_body(N) -> list:
    return N.body


def get_assign_target(N) -> str:
    targets = N.targets
    assert len(targets) == 1, "multiple targets in Assign not supported"
    return targets[0].id


def get_rhs(N):
    if is_assign(N):
        return N.value
    else:
        assert False


def get_lhs(N):
    if is_assign(N):
        targets = N.targets
        assert len(targets) == 1, "multiple targets in Assign not supported"
        return targets[0]
    elif is_aug_assign(N):
        return N.target
    else:
        return False
    
        
def is_rhs_name(N):
    return is_name(get_rhs(N))


def get_name_id(N):
    assert is_name(N)
    return N.id



import gc
import numpy as np

def list_all_numpy_objects():
    all_objects = gc.get_objects()  # Get all objects tracked by the garbage collector
    numpy_objects = [obj for obj in all_objects if isinstance(obj, np.ndarray)]  # Filter for numpy arrays
    return numpy_objects

def foo(a, r, b, iters):
    for i in range(iters):
        # Check the number of objects in memory before the operation
        before_count = len(gc.get_objects())
        
        c = 1.0 / (a @ b)  # Compute intermediate result
        
        # Check the number of objects in memory after the operation
        after_count = len(gc.get_objects())
        
        # Assign b to new value
        b = r @ c
        
        print(f"Iteration {i}: Objects before: {before_count}, Objects after: {after_count}")
    
    return b

# Example usage with dummy arrays
import numpy as np
a = np.random.rand(1000, 1000)
r = np.random.rand(1000, 1000)
b = np.random.rand(1000, 1000)
iters = 5

foo(a, r, b, iters)
print(list_all_numpy_objects())



import sys
import ast
from pprint import pprint
import traceback
import logging

from . import mytypes as types
from . import glb
from . import utils
from . import libfuncs
from .symboltable import symtab
from .glb import UnhandledNodeException
from .glb import dump
from .astutils import *
from . import utils
import defuse


class EscapeAnalysis(defuse.DefUseAnalysis):
    def __init__(self):
        defuse.DefUseAnalysis.__init__(self)
        self.symbols = None
        self.logger = utils.get_logger('ea')

        
    def visit_FunctionDef(self, F: ast.FunctionDef):
        self.visit_FunctionDef_main(F)
        self.symbols = symtab.get_func_symbol_table(F.name)
        self.do_function(F)


    def do_function(self, F: ast.FunctionDef):
        F.aliased_var = []
        for stmt in F.body:
            self.do_stmt(stmt, F)
            

    def do_stmt(self, s, F):        
        if is_assign(s):
            rhs = get_rhs(s)
            if is_name(rhs):
                var = get_name_id(rhs)
                ty = symtab.get_local_type(F.name, var)
                if types.is_array(ty) and var not in F.aliased_var: 
                    F.aliased_var.append(var)
                    self.logger.debug("add aliased var: %s" % var)

        if is_loop(s):
            self.do_loop(s, F)

            
    def get_loop_private_vars(self, L):
        if is_for(L):
            defuse.DefUseAnalysis.visit_For(self, L)
        if is_while(L):
            defuse.DefUseAnalysis.visit_While(self, L)
        myinfo = self.level_tree_current
        vars_local = (myinfo.vars_may_def | myinfo.vars_cond_def) - myinfo.vars_use
        return vars_local
    
        
    def do_loop(self, L, F):
        privatevars = self.get_loop_private_vars(L)
        #print(privatevars)
        body = get_loop_body(L)
        sites = []
        L.alloc_sites = sites
        
        for stmt in body:
            self.do_stmt(stmt, F)
            if is_alloc_site(stmt):
                lhs = get_lhs(stmt)
                assert is_name(lhs)
                var = get_name_id(lhs)
                # if var == 'frontier':
                #     continue
                # if var == 'r':
                #     continue
                if var in F.aliased_var:
                    self.logger.debug('skip aliased var: %s' % var)
                    continue
                if var not in privatevars:
                    self.logger.debug('skip nonprivate var: %s' % var)
                    continue
                sites.append(stmt)


    # def do_stmt(self, s, F):
    #     # Check defs
    #     if is_local_def(s):
    #         var = get_defined_name(s)
    #         self.symbols.add_def(var, s)

    #     # Check uses    
    #     if is_setitem(s):
    #         var = get_setitem_base(s)
    #         self.symbols.add_use(var, s)
    #         value = get_setitem_value(s)
    #         if is_name(value):
    #             self.symbols.add_use(value.id, s)
    #         elif is_bin_op(value):
    #             for operand in get_bin_op_operands(value):
    #                 if is_name(operand):
    #                     self.symbols.add_use(operand.id, s)
    #         else:
    #             print('[unhandled]:', value)
        
    #     if is_loop(s):
    #         self.do_loop(s, F)

         



import numpy as np
import intrepydd as pydd
from intrepydd.lang import *
import inspect

def sinkhorm_wmd(K: Array(float64, 2), M: Array(float64, 2), r: Array(float64, 1), x: Array(float64, 2), max_iter: int32, values: Array(float64, 1), columns: Array(int32, 1), indexes: Array(int32, 1), ncols_c: int32) -> Array(float64, 1):
    c = csr_to_spm(values, columns, indexes, ncols_c)
    it = 0
    while it < max_iter:
        u = 1.0 / x
        v = c.spm_mul(div(1.0, K.T @ u))
        x = spmm_dense(div(1.0, r).mul(K), v)
        r[0] = 0

        it += 1

    u = 1.0 / x    
    v = c.spm_mul(div(1.0, K.T @ u))
    return mul(u, spmm_dense(mul(K, M), v)).sum(0)

cppcode, pycode = pydd.compile_from_src(inspect.getsource(sinkhorm_wmd), dumppy=True, licm=True, dense_array_opt=True, sparse_array_opt=True)
print(cppcode)


from . import libfuncs
from .utils import deprecated

class CallSig(object):
    def __init__(self, module, funcname, arg_types):
        self.module = module
        self.funcname = funcname
        self.arg_types = arg_types

    def __str__(self):
        s = 'call signature: %s::%s' % (self.module, self.funcname)
        return s


class FuncSymbolTable(object):
    '''
    Represents symbol table for a function
    '''
    def __init__(self):
        self.defs = {}
        self.uses = {}
        self.locals = []
        self.unboxed = {}
        

    def register_local(self, var):
        self.locals.append(var)


    def add_def(self, name, stmt):
        if name not in self.defs:
            self.defs[name] = []
        self.defs[name].append(stmt)    
        

    def add_use(self, name, stmt):
        if name not in self.uses:
            self.uses[name] = []
        self.uses[name].append(stmt)    
        

    def register_unboxed_array(self, var, info):
        self.unboxed[var] = info


    def is_unboxed(self, var):
        return var in self.unboxed

        
    def get_unboxed_ptr(self, var):
        return self.unboxed[var]
        

class SymbolTable(object):
    '''
    Each module has a symbol table.
    '''
    def __init__(self):
        self.user_globals = {}
        self.func_table = {}

    # def init_typemap(self):
    #     self.typemap = {} # for user defined functions

    # def register_type(self, name, ty):
    #     self.typemap[name] = ty

    def register_user_func(self, F):
        name = F.name
        self.user_globals[name] = F.ret_type
        self.func_table[name] = F.localtypes
        

    def get_local_type(self, funcname, name):
        assert type(funcname) == str, type(funcname)
        assert type(name) == str, type(name)
        table = self.func_table[funcname]
        msg = "local %s not found in %s" % (name, funcname)
        assert name in table, msg
        return table[name]
        

    @deprecated("Please use the next `get_type` below")
    def get_type(self, callSig: CallSig):
        name = callSig.funcname
        module = callSig.module

        if module == '': # this module            
            if name in self.user_globals:
                return self.user_globals[name]
            else:
                raise glb.UndefinedSymbolException(module, name)
        else: # either a global or a library function        
            return libfuncs.get_ret_type(callSig)


    def get_type(self, name, module, type_args=None):
        if module == '': # this module            
            if name in self.user_globals:
                return self.user_globals[name]
            else:
                raise glb.UndefinedSymbolException(module, name)
        else: # either a global or a library function        
            return libfuncs.get_type(name, module, type_args)


    def is_user_defined_symbol(self, name):
        return name in self.user_globals


    def get_func_symbol_table(self, func):
        return self.func_table[func]

    
    def register_unboxed_arrays(self, func, name):
        pass

    
    def __str__(self):
        s = 'User defined functions are:\n'
        for name in self.typemap:
            s += name + ' => ' + str(self.typemap[name]) + '\n'
        return s

    

symtab = SymbolTable()



import numpy as np
import timeit
import intrepydd
from intrepydd.lang import *

def foo(a: Array(float64, 2), b: Array(float64, 2), c: Array(float64, 2)) -> Array(float64, 2):
    return c / (a @ b)


foo1 = intrepydd.compile(foo)

N = 2000
a = np.random.randn(N, N)
b = np.random.randn(N, N)
c = np.random.randn(N, N)
assert np.allclose(foo(a, b, c), foo1(a, b, c))
print(timeit.timeit(lambda: foo(a, b, c), number=10) / 10)
print(timeit.timeit(lambda: foo1(a, b, c), number=10) / 10)


#!/usr/bin/env python3

import ast
import sys
import os
from .typeinfer import TypeInferer
from .codegen import ModuleGen
from .scanner import Scanner
from .transform1 import TypeFreeTransformer
from .transform2 import TypedTransformer
from .ret_to_assign import ReturnExprToStmt
from . import glb


def get_src_as_str():
    output = ''
    for line in open(glb.args.file):
        sline = line.strip()
        if sline.startswith('pfor ') and sline[-1] == ':':
            sline = line.replace('pfor ', 'for ')
            line = sline.replace('\n', ' # type: pfor\n')
     
        output += line
    #print(output)        
    return output

def make_python_code(code):
    newcode = ''
    newcode += 'from pydd_debug import *'
    return

def gen_python_code(code):
    # debug_path = os.path.dirname(os.path.realpath(__file__)) + '/debugging'
    # prefix = 'import sys\n'
    # prefix += 'sys.path.append("%s")\n' % debug_path
    # prefix += 'from pydd_types import *\n'
    # prefix += 'from pydd_python import *\n'
    # prefix += 'from pydd_numpy import *\n'
    prefix = 'from intrepydd.lang import *\n\n'
    print(prefix+code, file=open(glb.get_filepath()+glb.get_module_name()+'.py', 'w'))

def gen_numba_code(code):
    print("Numba code generation is not supported yet!")
    sys.exit(0)

def gen_cpp_module(code):
    '''
    Each pydd file goes through three passes:

    - Scanner
      - Does some preprocessing stuff

    - TypeInferer
      - Does type inference

    - ModuleGen
      - Does the actul translation and generate the cpp module

    - TypeFreeTransformer
      - Apply AST transformations independent from type information

    - TypedTransformer
      - Apply AST transformations based on type information
    '''
    
    glb.open_cpp_module(glb.get_filepath()+glb.get_module_name())
    tree = ast.parse(code)
    if glb.args.verbose:
        print("AST parsing done.")

    tree = ReturnExprToStmt().visit(tree)

    scanner = Scanner()
    scanner.visit(tree)
    if glb.args.verbose:
        print("Scanner phase done.")        

    transform1 = TypeFreeTransformer()
    transform1.visit(tree)
    if glb.args.verbose:
        print("Transform1 phase done.")        

    analyzer = TypeInferer()
    analyzer.visit(tree)
    if glb.args.verbose:
        print("Type analyze phase done.")        

    transform2 = TypedTransformer()
    transform2.visit(tree)
    if glb.args.verbose:
        print("Transform2 phase done.")

    # This code is potentially buggy, to fix
    # if glb.args.alloc_opt:
    #     import escapeanalysis
    #     import escapetransform
    #     # This is hacky and inefficient but works for now 
    #     for _ in range(3):
    #         ea = escapeanalysis.EscapeAnalysis()
    #         ea.visit(tree)
    #         rewriter = escapetransform.AllocHoistTransformer()
    #         rewriter.visit(tree)

    if glb.args.g:
        import debugtransform
        rewriter = debugtransform.DebugTransformer()
        rewriter.visit(tree)

    if 'dumppy' in glb.options and glb.options['dumppy']:
        gen_python_code(ast.unparse(tree))

    module = ModuleGen()
    module.visit(tree)
    if glb.args.verbose:
        print("Code gen phase done.")

    glb.close_cpp_module()
        
def main():    
    # with open(f, "r") as source:
    #     tree = ast.parse(source.read())
    code = get_src_as_str()
    if glb.args.opt_level == 0:
        gen_python_code(code)
    elif glb.args.opt_level == 1:
        gen_numba_code(code)
    elif glb.args.opt_level == 2:
        if glb.args.file.find('.cpp') == -1:
            gen_cpp_module(code)
        
        
    
        
if __name__ == "__main__":
    main()




import numpy as np
import intrepydd
from intrepydd.lang import *

def foo(a: Array(float64, 1), b: Array(float64, 1)):
    x = 0.5 * a + b
    return x

foo1 = intrepydd.compile(foo, dense_array_opt=True)

N = 100000
a = np.random.randn(N)
b = np.random.randn(N)
assert np.allclose(foo(a, b), foo1(a, b))


import os
import sys
from pathlib import Path
import inspect
import shutil
import hashlib
import importlib
import subprocess
from . import glb, launcher

def install_and_import(package):
    try:
        # Check if the package is already installed
        m = importlib.import_module(package)
        #print(f"'{package}' is already installed.")
    except ImportError:
        # Package not found, install it
        print(f"'{package}' not found. Installing...")
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])
        print(f"'{package}' has been installed.")
        m = importlib.import_module(package)
    return m


def compile_from_file(file, args):
    sys.argv = sys.argv[:1]  # to make it work when hit "compile" multiple times in the web version
    sys.argv += [file] + args + ['-dumppy'] #+ ["-v"]
    glb.init()
    
    #os.chdir(glb.get_filepath())
    launcher.main()

def compile_from_src(src, no_cfg=False, dense_array_opt=False, sparse_array_opt=False, licm=False, slice_opt=False, dumppy=False):
    import datetime
    dir = '/tmp/'
    filename = dir + f'kernel_{hashlib.sha256(src.encode()).hexdigest()}.pydd'
    Path(filename).write_text(src)
    args = []
    if dense_array_opt:
        args.append('-dense-opt')
    if sparse_array_opt:
        args.append('-sparse-opt')
    if licm:
        args.append('-licm')
    if slice_opt:
        args.append('-slice_opt')
    if dumppy:
        args.append('-dumppy')
    compile_from_file(filename, args)
    cppcode = Path(filename.replace('.pydd', '.cpp')).read_text()
    if no_cfg and '%>\n' in cppcode:
        cppcode = cppcode.split('%>\n')[1].strip()
    if dumppy:
        pycode = Path(filename.replace('.pydd', '.py')).read_text()
        return cppcode, pycode
    else:
        return cppcode

def compile(fn, print_cpp=False, dense_array_opt=False, sparse_array_opt=False, licm=False, slice_opt=False):
    source_code = inspect.getsource(fn)
    cpp_code = compile_from_src(source_code, 
                    dense_array_opt=dense_array_opt, 
                    sparse_array_opt=sparse_array_opt, 
                    licm=licm, 
                    slice_opt=slice_opt)
    if print_cpp:
        print(cpp_code)
    module_name = ''
    for line in cpp_code.split():
        if line.startswith('PYBIND11_PLUGIN('):
            module_name = line.split('PYBIND11_PLUGIN(')[1].split(')')[0]
            break
    assert module_name != ''
    
    dir = '/tmp/'
    cpp_file = dir + module_name + '.cpp'
    copied_file = module_name + '.cpp'
    shutil.copy(cpp_file, copied_file)
    cppimport = install_and_import('cppimport')
    module = cppimport.imp(module_name)
    Path(copied_file).unlink()
    Path('.rendered.' + copied_file).unlink(missing_ok=True)
    return getattr(module, fn.__name__)


import sys
import ast
from pprint import pprint
from . import mytypes
from . import glb
from . import libfuncs
from . import astutils
from .glb import UnhandledNodeException
from .symboltable import symtab
from .licmarray import get_num_array_dim
from .utils import deprecated
from . import utils


def dump(N):
    print(ast.dump(N))

class ModuleGen(ast.NodeVisitor):
    # WARN: name can be a full path
    def __init__(self):
        self.name = glb.get_module_name()
        self.funcs = []
        self.cur_func = None
        self.cur_node = None        
        self.inv_subs = set()
        self.has_main = False
        self.type_logger = utils.get_logger('type')
        

    def writeln(self, s=''):
        glb.cpp_module.writeln(s)

    def write(self, s):
        glb.cpp_module.write(s)

    def visit_Module(self, node):
        module = glb.get_module_name()
        # Gen cppimport headers (compiler flags etc)
        if glb.args.host == 'py':
            glb.cpp_module.do_pymodule_header()
        # Gen default includes
        glb.cpp_module.do_default_include()
        
        # Gen functions declarations
        glb.cpp_module.writeln("namespace %s {" % module)
        glb.cpp_module.writeln("/* Function declarations */")
        for f in glb.cpp_module.funcs:
            self.gen_func_header(f)
        glb.cpp_module.writeln("")
        # Gen function definitions
            
        self.generic_visit(node)
        glb.cpp_module.writeln("}") # namespace ends    
        # Gen cppimport footer
        if glb.args.host == 'py':    
            glb.cpp_module.do_py_footer()

        if self.has_main and glb.args.host == 'cpp':
            self.gen_real_main()
    
        
    def visit_Import(self, node):
        self.generic_visit(node)


    def dump_local_type(self, N):
        if not glb.args.dump_type:
            return
        print('=== local types for function "%s" ===' % N.name)
        typemap = symtab.get_func_symbol_table(N.name)
        if glb.args.dump_type == '*':
            for var in typemap:
                print("    %s: %s" % (var, typemap[var]))
        

    def visit_FunctionDef(self, N):
        self.dump_local_type(N)
        self.cur_func = N
        if N.name == 'main':
            self.has_main = True
        self.register(N)
        
        self.do_func_header(N)
        self.do_var_decls(N)
        # self.do_func_args(N)
        self.do_func_body(N)
        self.writeln('}\n')
        #self.generic_visit(node)

        
    def register(self, N):
        self.funcs.append(N)


    def gen_debug_info(self, N):
        if not glb.args.g:
            return
        s = 'pydd::append_to_call_stack("%s");' % N.name
        self.writeln(s)
        

    def gen_real_main(self):
        s = '''
        int main(int argc, void** argv) { %s::main(); }
        ''' % glb.get_module_name()
        self.writeln(s)

    def gen_func_header(self, N: ast.FunctionDef):
        s = '%s %s(' % (N.ret_type.get_declare_str(), N.name)
        
        args = []
        for arg in N.args.args:            
            args.append(
                arg.type.get_declare_str() + " " + arg.arg
                )
        s += ', '.join(args) + ');\n'
        glb.cpp_module.write(s)
            

    def do_var_decls(self, N: ast.FunctionDef):
        self.writeln('/* Declarations */')
        args = []
        
        for arg in N.args.args:
            args.append(arg.arg)

        s = ''     
        for v in N.typemap:
            if v in args:
                continue
            ty = N.typemap[v]
            s += ty.get_declare_str() + ' ' + v + ';\n'

        arraydims = self.get_array_dim_info(N)
        if 'licm_array' in glb.options and glb.options['licm_array']:
            if hasattr(N, 'inv_array_dim'):
                for array_id in N.inv_array_dim:
                    #print('array', array_id, 'dim', N.inv_array_dim[array_id])
                    if array_id in args:
                        continue
                    self.gen_data_ptr_decls(array_id, N.inv_array_dim[array_id])
            
            for v in args:
                ty = self.get_local_type(v)
            # for v in N.typemap:
            #     ty = self.get_local_type(v)
                if isinstance(ty, mytypes.NpArray):
                    etype = ty.etype.get_cpp_name()
                    s += '%s* %s_data_ptr_pydd = %s.mutable_data();\n' \
                        % (etype, v, v)
                    if ty.ndim == 2:    
                        s += 'int %s_shape1_pydd = %s.shape(1);\n' \
                            % (v, v)
                    
        self.writeln(s+'\n')

    
    def get_array_dim_info(self, F):
        if hasattr(F, 'inv_array_dim'):
            return F.inv_array_dim
        else:
            return {}

        
    @deprecated('')
    def gen_data_ptr(self, name):
        assert type(name) == str, type(name)
        ty = self.get_local_type(name)
        if not isinstance(ty, mytypes.NpArray):
            return

        s = ''
        data_ptr = name + '_data_ptr_pydd'
        arr_shape1 = name + '_shape1_pydd'

        s += '%s* %s = %s.mutable_data();\n' % (ty.get_elt_type(), data_ptr, name)
        s += '%s %s = %s.shape(1);\n' % (mytypes.int64, arr_shape1, name)
        self.writeln(s)


    def gen_data_ptr_decls(self, name, ndim):
        assert type(name) == str, type(name)
        ty = self.get_local_type(name)
        if not isinstance(ty, mytypes.NpArray):
            return

        data_ptr = name + '_data_ptr_pydd'
        s = '%s* %s;\n' % (ty.get_elt_type(), data_ptr)
        for d in range(1, ndim):
            arr_shape_d = '%s_shape%i_pydd' % (name, d)
            s += '%s %s;\n' % (mytypes.int64, arr_shape_d)
        self.write(s)

        
    def gen_data_ptr_assign(self, name, ndim):
        assert type(name) == str, type(name)

        data_ptr = name + '_data_ptr_pydd'
        s = '%s = %s.mutable_data();\n' % (data_ptr, name)
        for d in range(1, ndim):
            arr_shape_d = '%s_shape%i_pydd' % (name, d)
            s += '%s = %s.shape(%i);\n' % (arr_shape_d, name, d)
        return s

    
    def do_func_header(self, N: ast.FunctionDef):
        s = '%s %s(' % (N.ret_type.get_declare_str(), N.name)
        
        args = []
        for arg in N.args.args:            
            args.append(
                arg.type.get_declare_str() + " " + arg.arg
                )
        s += ', '.join(args) + ')'
        self.write(s)
        self.writeln(' {')
        
            
    def do_func_body(self, N: ast.FunctionDef):
        for stmt in N.body:
            if self.should_ignore(stmt):
                continue

            self.cur_node = stmt
            
            s = self.eval_node(stmt)
            #print('stmt: ', s)
            self.writeln(s + ';')

    def is_str_stmt(self, N):
        if isinstance(N, ast.Expr):
            if isinstance(N.value, ast.Str):
                return True
        return False    

    def should_ignore(self, stmt):
        if self.is_str_stmt(stmt):
            return True
        return False

    def eval(self, N):
        return self.eval_node(N)

    def eval_node(self, N):
        s = ''
        if isinstance(N, str): # a little hacky
            s = N 
        elif isinstance(N, ast.Assign):
            s = self.eval_Assign(N)
        elif isinstance(N, ast.AnnAssign):
            s = self.eval_AnnAssign(N)
        elif isinstance(N, ast.Name):
            s = self.eval_name(N)
        elif isinstance(N, ast.Num):
            s = self.eval_Num(N)
        elif isinstance(N, ast.For):
            s = self.eval_For(N)
        elif isinstance(N, ast.While):
            s = self.eval_While(N)
        elif isinstance(N, ast.If):
            s = self.eval_if(N)
        elif isinstance(N, ast.Return):
            s = self.eval_Return(N)
        elif isinstance(N, ast.NameConstant):
            s = self.eval_name_constant(N)
        elif isinstance(N, ast.List):
            s = self.eval_List(N)
        elif isinstance(N, ast.Tuple):
            s = self.eval_Tuple(N)
        elif isinstance(N, ast.Expr):
            s = self.eval_Expr(N)
        elif isinstance(N, ast.BoolOp):
            s = self.eval_BoolOp(N)
        elif isinstance(N, ast.Str):
            s = self.eval_Str(N)
        elif isinstance(N, ast.Subscript):
            s = self.eval_Subscript(N)
        elif isinstance(N, ast.Dict):
            s = self.eval_Dict(N)
        elif isinstance(N, ast.Call):
            s = self.eval_Call(N)
        elif isinstance(N, ast.Compare):
            s = self.eval_Compare(N)
        elif isinstance(N, ast.UnaryOp):
            s = self.eval_UnaryOp(N)    
        elif isinstance(N, ast.BinOp):
            s = self.eval_BinOp(N)    
        elif isinstance(N, ast.Continue):
            s = self.eval_Continue(N)
        elif isinstance(N, ast.Break):
            s = self.eval_Break(N)
        elif isinstance(N, ast.Pass):
            s = self.eval_Pass(N)
        elif isinstance(N, ast.AugAssign):
            s = self.eval_AugAssign(N)
        elif isinstance(N, ast.Attribute):
            s = self.eval_Attribute(N)
        elif isinstance(N, ast.Assert):
            s = self.eval_Assert(N)
        else:
            glb.exit_on_unsupported_node(N)
        return s
                
    def get_type(self, N):
        if not hasattr(N, 'type'):
            self.do_node(N)
        return a.type
            
    def eval_name(self, N):
        return N.id


    def eval_AugAssign(self, N:ast.AugAssign):
        N.targets = [N.target]
        N.value = self.eval_BinOp_raw(N.op, N.target, N.value)
        return self.eval_Assign(N)

    def eval_AnnAssign(self, N):
        N.targets = [N.target]
        return self.eval_Assign(N)

    def is_array_slice(self, N):
        if isinstance(N, ast.Call) and N.func.id == "get_row":
            return True
        return False
    
    def eval_Assign(self, N: ast.Assign):
        assigns = []        
        for target in N.targets:
            # Assign to one var
            if isinstance(target, ast.Name):
                lhs = target.id
                rhs = N.value
                # if glb.args.slice_opt and self.is_array_slice(rhs):
                #     # R.H.S is array slice, special codegen may be needed
                #     s = self.gen_unboxed_slice(target, N.value)                    
                # else:
                
                s = '%s = %s' % (target.id, self.eval(N.value))    
                assigns.append(s)
                
            # Assign to multiple vars    
            elif isinstance(target, ast.Tuple):
                var_tp = target.elts
                value_tp = N.value.elts
                assert len(var_tp) == len(value_tp)
                for i in range(len(var_tp)):
                    s = '%s = %s' % (var_tp[i].id, self.eval(value_tp[i]))
                    assigns.append(s)
            # Assign to subscript
            # Need some interpretation
            elif isinstance(target, ast.Subscript):
                #s = self.eval_get_or_set_item(self.eval(target.value), self.eval_Subscript_slice(target.slice), self.eval(N.value))
                s = self.eval_get_or_set_item(self.eval(target.value), target, self.eval(N.value))

                # s = "pydd::setitem(%s, %s, %s)" % \
                #   (self.eval(target.value), self.eval(N.value),
                #    self.eval_Subscript_slice(target.slice))
                assigns.append(s)
                
            else:
                glb.exit_on_unsupported_node(N)
            
        #print(assigns)        
        return '; '.join(assigns)


    def get_call_arg_str(self, call, i):
        assert isinstance(call, ast.Call)
        args = call.args
        arg = args[i]
        if isinstance(arg, ast.Name):
            return arg.id
        
    
    def get_data_type(self, slice_call):
        assert isinstance(slice_call, ast.Call)
        funcid = slice_call.func.id
        args = slice_call.args
        if funcid == 'get_row':
            ty = self.get_local_type(args[0].id)
            return ty.etype.get_cpp_name()
        else:
            assert False, "get_data_type only supports get_row call"
        

    def gen_unboxed_slice(self, N):
        s = ''
        base = self.get_call_arg_str(N, 0)
        index = self.get_call_arg_str(N, 1)
        s = '%s_data_ptr_pydd + %s_shape1_pydd * %s' % (base, base, index)
        return s
    
    # def gen_unboxed_slice(self, target, value):
    #     s = ''
    #     lhs = target.id
    #     table = symtab.get_func_symbol_table(self.cur_func.name)
    #     table.register_unboxed_array(lhs, (lhs+"_data"))
    #     ty = self.get_data_type(value)
    #     base = self.get_call_arg_str(value, 0)
    #     index = self.get_call_arg_str(value, 1)
    #     s = '%s* %s_%s_data = %s_data_ptr_pydd + %s_shape1_pydd * %s' \
    #                     % (ty, base, index, base, base, index)
    #     table.register_unboxed_array(lhs, '%s_%s_data' % (base, index))
    #     return s
        

    def eval_Num(self, N: ast.Num):
#        dump(N)
        #return '(%s)%s' % (N.type, str(N.n))
        return '%s' % (str(N.n))
    
        
    def eval_name_constant(self, N: ast.Num):
        if isinstance(N.value, bool):
            if N.value is True:
                return "true"
            else:
                return "false"
        else:
            glb.exit_on_unsupported_node(N)

            
    def eval_While(self, N):
        s = ''
        #print('while at line %d inv set:' % astutils.get_line_number(N))
        if hasattr(N, 'inv_subscripts'):
            seen = set()
            for sub in N.inv_subscripts:
                name = sub.value.id
                ty = self.get_local_type(name)
                if isinstance(ty, mytypes.NpArray):
                    self.inv_subs.add(sub)
                    #print(name)
                    if name not in seen:
                        seen.add(name)
                        ndim = get_num_array_dim(sub)
                        assert ndim > 0
                        s += self.gen_data_ptr_assign(name, ndim)
            #print('')

        cond = self.eval(N.test)
        s += 'while (%s) {\n' % cond
        for expr in N.body:
            s += self.eval(expr) + ';\n'
        s += '}'
        return s

    
    def is_dict(self, N):
        return isinstance(N.type, mytypes.DictType)

    
    def eval_For(self, N: ast.For):
        s = ''
        #print('for at line %d inv set:' % astutils.get_line_number(N))
        if hasattr(N, 'inv_subscripts'):
            seen = set()
            for sub in N.inv_subscripts:
                name = sub.value.id
                ty = self.get_local_type(name)
                if isinstance(ty, mytypes.NpArray):
                    self.inv_subs.add(sub)
                    #print(name)
                    if name not in seen:
                        seen.add(name)
                        ndim = get_num_array_dim(sub)
                        assert ndim > 0
                        s += self.gen_data_ptr_assign(name, ndim)
            #print('')

        if N.is_pfor:
            libfuncs.used_packages += ['shared/omp']
            s += '#pragma omp parallel for'
            if N.vars_private:
                s += ' private(%s)' % N.vars_private
            if N.vars_lastprivate:
                s += ' lastprivate(%s)' % N.vars_lastprivate
            s += '\n'
        if self.is_dict(N.iter):
            s += self.eval_For_Dict(N)
        else:
            s += self.eval_For_Array(N)        
        
        return s

    def eval_For_Dict(self, N):
        target = self.eval_node(N.target)
        iter = self.eval_node(N.iter)
        start = '%s->begin()' % iter
        end = '%s->end()' % iter
        index = '_it'

        s = 'for (%s::iterator %s = %s; %s != %s; %s++) {\n' % \
          (N.iter.type.get_cpp_name(), index, start, index, end, index)
        s += '%s = %s->first;\n' % (target, index)
        for expr in N.body:
            s += self.eval_node(expr) + ';\n'
        s += '}'
        return s


    def eval_For_Array(self, N):
        target = self.eval_node(N.target)
        iter = self.eval_node(N.iter)

        start = '0'
        end = 'pydd::len(%s)' % iter
        step = '1'
        index = '_i'

        if self.is_range_for(N):
            start, end, step = self.get_range_args(N.iter)
        
        s = 'for (int %s = %s; %s < %s; %s += %s) {\n' % (index, start, index, end, index, step)
        if self.is_range_for(N):
            s += '%s = %s;\n' % (target, index)
        else:
            if glb.args.slice_opt and astutils.is_name(iter):
                s += '%s = pydd::getitem%s(%s, %s);\n' % (target, self.get_dimension_suffix(iter), iter, index)
            else:
                s += '%s = pydd::getitem(%s, %s);\n' % (target, iter, index)
        for expr in N.body:
            s += self.eval_node(expr) + ';\n'
        s += '}'
        return s

    def is_range_for(self, N: ast.For):
        if isinstance(N.iter, ast.Call):
            if N.iter.func.id == 'range':
                return True

        return False

    def get_range_args(self, N: ast.Call):
        if len(N.args) == 1:
            return (0, self.eval(N.args[0]), 1)
        elif len(N.args) == 2:
            return (self.eval(N.args[0]), self.eval(N.args[1]), 1)
        elif len(N.args) == 3:
            return (self.eval(N.args[0]), self.eval(N.args[1]), self.eval(N.args[2]))
        else:
            assert False
        return

    def eval_if(self, N: ast.If):
        '''        
        if 'test':
          'body'
        'orelse':

        '''
        test = self.eval(N.test)
        s = 'if (%s) {\n' % test
        for expr in N.body:
            s += self.eval(expr) + ';\n'
        s += '} '

        if not N.orelse:
            return s

        s += 'else '
        if isinstance(N.orelse[0], ast.If):
            assert len(N.orelse) == 1
            s += self.eval_if(N.orelse[0])
        else:
            s += '{\n'
            for expr in N.orelse:
                s += self.eval(expr) + ';\n'
            s += '}'    
        return s

    
    def get_local_type(self, name):
        if name not in self.cur_func.typemap:
            assert False, 'name not found:' + name
        return self.cur_func.typemap[name]


    def get_dimension_suffix(self, var):
        ty = self.get_local_type(var)
        call = ''
        if ty.get_dimension() == 1:
            call += '_1d'
        elif ty.get_dimension() == 2:
            call += '_2d'
        else:
            pass
        return call    

    
    def eval_get_or_set_item(self, base, sub, value=None):
        indices = self.eval_Subscript_slice(sub.slice)
        assert isinstance(indices, list)
        basetype = self.get_local_type(base)
        call = 'getitem'
        args = []

        if value != None:
            call = 'setitem'
        args.append(base)

        # if sub not in self.inv_subs:
        #     print(base)
        #     print('sub:', ast.dump(sub))
        #     print(basetype.get_dimension())

        ## This is missing _data_ptr assigns
        # if basetype.get_dimension() > 0:
        #     #ndim = get_num_array_dim(sub)
        #     ndim = basetype.get_dimension()
        #     if ndim == 1:
        #         call += '_1d'
        #         args[0] = base+'_data_ptr_pydd'
        #     elif ndim == 2:
        #         call += '_2d'
        #         args[0] = base+'_data_ptr_pydd'
        #         args.append(base+'_shape1_pydd')
        #     if sub in self.inv_subs:    
        #         self.inv_subs.remove(sub)
            
        if sub in self.inv_subs:
            ndim = get_num_array_dim(sub)
            if ndim == 1:
                call += '_1d'
                args[0] = base+'_data_ptr_pydd'
            elif ndim == 2:
                call += '_2d'
                args[0] = base+'_data_ptr_pydd'
                args.append(base+'_shape1_pydd')
                
            self.inv_subs.remove(sub)    
        elif glb.args.unchecked_access and isinstance(basetype, mytypes.NpArray):
            if basetype.get_dimension() == 1:
                call += '_1d'
            elif basetype.get_dimension() == 2:
                call += '_2d'
        else:
            # -- This is getitem but is not array, like a dict getitem
            pass

        
        if value:
            args.append(value)
            
        args += indices
        s = "pydd::%s(%s)" % (call, ', '.join(args))
        return s
        

    def eval_Subscript(self, N: ast.Subscript):
        '''
        Fields: N.value[N.slice]
        '''
        # basetype = self.get_local_type(N.value.id)
        # call = 'getitem'
        # if isinstance(basetype, mytypes.NpArray):
        #     if basetype.get_dimension() == 1:
        #         call = 'getitem_1d'
        #     elif basetype.get_dimension() == 2:
        #         call = 'getitem_2d'
        # s = "pydd::%s(%s, %s)" % \
        #           (call, self.eval(N.value), self.eval_Subscript_slice(N.slice))
        # return self.eval_get_or_set_item(N.value.id, self.eval_Subscript_slice(N.slice))
        return self.eval_get_or_set_item(N.value.id, N)

    def eval_Subscript_slice(self, v):
        '''
        a[0] returns ['0']
        #a[0,1] returns '{0, 1}'
        a[0,1] returns ['0', '1']
        '''
        indices = []
        if isinstance(v, ast.Tuple):
            for e in v.elts:
                indices.append(self.eval(e))            
        else:
            indices.append(self.eval(v))
        return indices


    def eval_Dict(self, N: ast.Dict):
        assert len(N.keys) == 0, "Support for initialized dict is \
under development, please create an empty Dict and then add elements \
for now"

        # s = 'new std::map<%s, %s>{' % (N.type.key, N.type.value)
        s = 'new std::unordered_map<%s, %s>{' % (N.type.key, N.type.value)
        s += '}'
        return s

    def eval_Str(self, N):
        return '"%s"' % N.s
    
    def eval_Expr(self, N: ast.Expr):
        if isinstance(N.value, ast.Call):
            return self.eval_Call(N.value)
        else:
            raise glb.exit_on_unsupported_node(N)

    def eval_Call(self, N: ast.Call):
        # eg: Call(func=Name(id='print', ctx=Load()), args=[Name(id='i', ctx=Load())], keywords=[])
        # todo
        
        funcname  = ''
        ns = ''
        # Get function name and namespace if any
        if isinstance(N.func, ast.Name):
            funcname = N.func.id
            # if funcname == 'sub':
            #     print("funcname", funcname, 'in', self.cur_func.name)
            if glb.is_user_defined_func(funcname):
                ns = ''
            elif funcname.startswith('_') or libfuncs.is_global_func(funcname):
                ns = 'pydd'
            else:
                glb.exit_on_node(N, 'Unrecognized function: '+funcname)
            
        elif isinstance(N.func, ast.Attribute):
            if hasattr(N, 'is_module_call'):
                ns = glb.cpp_module.get_module_namespace(N.func.value.id)
                funcname = N.func.attr
            else:
                glb.exit_on_node(N, "Attribute call not supported")


        # Generate the actual call and args
        if funcname == 'get_row_ptr':
            return self.gen_unboxed_slice(N)
        s = ''
        if ns:  
            s = ns + '::' + funcname
        else:
            s = funcname
   
        s += '('
        args = []
        for arg in N.args:
            # if isinstance(arg, ast.List):
            #     args.append(self.eval_ListArg(arg))
            if isinstance(arg, ast.Tuple):
                args.append(self.eval_ListArg(arg))
            else:
                realarg = self.eval(arg)
                # Func name ends with 1 implies using slice opt
                # if funcname.endswith('1'):
                    # name = arg.id
                    # table = symtab.get_func_symbol_table(self.cur_func.name)
                    # if table.is_unboxed(name):                        
                    #     realarg = table.get_unboxed_ptr(name)
                args.append(realarg)

        # if ns != '' and libfuncs.get_version_count(funcname, ns) > 1:
        #     args.append(str(glb.get_func_version(funcname))) 
       
        s += ', '.join(args)
        s += ')'
        return s

    def eval_Attribute(self, N):
        s = 'pydd::getattr_%s(%s)' % (N.attr, self.eval(N.value))
        return s

    # def is_builtin_func(self, N):
        
    #     if isinstance(N.func, ast.Name):
    #         funcname = N.func.id
    #         for F in self.funcs:
    #             if funcname == F.name:
    #                 return False
    #         return True
    #         # if N.func.id == 'print':
    #         #     return True
    #         # elif N.func.id == 'len':
    #         #     return True
    #     else:
    #         assert False
        
    #     return False
        
    def eval_List(self, N: ast.List):
        s = 'new std::vector<%s>{' % N.type.etype    
        s += ', '.join(list(map(lambda x: self.eval_node(x), N.elts)))
        s += '}'
        return s

    def eval_ListArg(self, N):
        s = '{'
        s += ', '.join(list(map(lambda x: self.eval_node(x), N.elts)))
        s += '}'
        return s
        
    def eval_Compare(self, N: ast.Compare):
        assert len(N.ops) == len(N.comparators)
        compares = []
        l = N.left
        for i in range(len(N.ops)):
            op = N.ops[i]
            r = N.comparators[i]
            compares.append(self.eval_cmpop(op, l, r))
            l = r
        return '&&'.join(compares)    

    def eval_cmpop(self, op, l, r):
        s = ''
        lhs = self.eval(l)
        rhs = self.eval(r)
        if isinstance(op, ast.Eq):
            s = '%s == %s' % (lhs, rhs)
        elif isinstance(op, ast.NotEq):
            s = '%s != %s' % (lhs, rhs)
        elif isinstance(op, ast.Lt):
            s = '%s < %s' % (lhs, rhs)
        elif isinstance(op, ast.LtE):
            s = '%s <= %s' % (lhs, rhs) 
        elif isinstance(op, ast.Gt):
            s = '%s > %s' % (lhs, rhs)
        elif isinstance(op, ast.GtE):
            s = '%s >= %s' % (lhs, rhs)
        else:
            raise UnhandledNodeException(op)

        return '(%s)' % s

    def eval_BoolOp(self, N):
        op = N.op
        opstr = ''
        s = ''
        if isinstance(op, ast.And):
            opstr = ' && '
            s = opstr.join(list(map(lambda x: self.eval(x), N.values)))
        elif isinstance(op, ast.Or):
            opstr = ' || '
            s = opstr.join(list(map(lambda x: self.eval(x), N.values)))
        else:
            glb.exit_on_unsupported_node(N)

        return s    
            

    def eval_BinOp(self, N: ast.BinOp):
        return '(' + self.eval_BinOp_raw(N.op, N.left, N.right) + ')'

    def eval_BinOp_raw(self, op, left, right):
        lhs = self.eval(left)
        rhs = self.eval(right)
        s = ''
        if isinstance(op, ast.Add):
            s += '%s + %s' % (lhs, rhs)
        elif isinstance(op, ast.Sub):
            s += '%s - %s' % (lhs, rhs)
        elif isinstance(op, ast.Mult):
            s += '%s * %s' % (lhs, rhs)
        elif isinstance(op, ast.Div):
            s += '((double)%s) / %s' % (lhs, rhs)
        elif isinstance(op, ast.FloorDiv):
            s += '%s / %s' % (lhs, rhs)
        elif isinstance(op, ast.Pow):
            s += 'pydd::pow(%s, %s)' % (lhs, rhs)
        elif isinstance(op, ast.LShift):
            s += '%s << %s' % (lhs, rhs)
        else:
            raise UnhandledNodeException('unrecognized op: '+op)
        return s

    def eval_UnaryOp(self, N):
        operand = self.eval(N.operand)
        s = ''
        if isinstance(N.op, ast.USub):
            s = '-%s' % operand 
        elif isinstance(N.op, ast.Not):
            s = '!%s' % operand 
        else:
            raise UnhandledNodeException(N)
        return s
            
    def eval_Continue(self, N: ast.Continue):        
        return 'continue'

    def eval_Break(self, N: ast.Break):        
        return 'break'

    def eval_Pass(self, N: ast.Break):        
        return '// pass'

    def eval_Return(self, N: ast.Return):        
        return 'return ' + self.eval(N.value)
        
    def eval_Assert(self, N: ast.Assign):
        
        return 'assert ' + self.eval(N.test)
        
        
    def visit_ImportFrom(self, node):
        self.generic_visit(node)

    def report(self):
        pprint(self.stats)




import numpy as np
import intrepydd
from intrepydd.lang import *

def foo(a: Array(float64, 1), b: Array(float64, 2), rows: List(int32)):
    for i in rows:
        a += get_row(b, i)
    return a


foo1 = intrepydd.compile(foo, print_cpp=True)

N = 10000
a = np.random.randn(N)
b = np.random.randn(N, N)
rows = np.random.randint(N, size=N//10)
assert np.allclose(foo(a, b, rows), foo1(a, b, rows))   